//------------------------------------------------------------------------------//
//             _________________________________________________                //
//            |                                                 |               //
//            |        Simple Delphi OpenGL Template 2.0        |               //
//            |_________________________________________________|               //
//                                                                              //
//                      Programmed by Alexandr Bezdek                         //
//                               -= 2017 =-                                     //
//                                                                              //
//------------------------------------------------------------------------------//
unit MyOpenGL;

interface

uses
  Winapi.Windows, Vcl.Controls, Vcl.ExtCtrls, System.Diagnostics, System.SysUtils,
  Vcl.Graphics, System.Math, dglOpenGL;

const
  cPI       : double = 3.1415926535897932384626433832795;
  cPIDiv180 : double = 0.01745329251994329576923690768489;
  c180DivPI : double = 57.295779513082320876798154814105;
  c2PI      : double = 6.283185307179586476925286766559;
  cInvPI    : double = 1/3.1415926535897932384626433832795;
  cInv2PI   : double = 1/6.283185307179586476925286766559;

type
// callback procedure for cyclic rendering driven by inner timer
  TCallBack = procedure(const DC : HDC; const Canvas : TControl) of object;

// line drawing styles
  TLineStyle = (lsSolid, lsDash, lsDashDot, lsDashDotDot, lsDot);

// units
  TVector3f           = record X, Y, Z : double; end;
  TVectorArray3f      = array of TVector3f;
  TVectorArrayArray3f = array of TVectorArray3f;
  TMatrix16f          = record X1, Y1, Z1, W1, X2, Y2, Z2, W2, X3, Y3, Z3, W3, X4, Y4, Z4, W4 : double; end;

// 3D object structure from pixels
  TPixelObject = record
    Structure : array of record
      Vertices  : TVectorArray3f;
      Color     : TColor;
      Alpha     : single;
      PixelSize : single;
    end;
    Position : TVector3f;
    Rotation : TVector3f;
    Scale    : TVector3f;
    Visible  : boolean;
  end;
  TPixelObjectArray = array of TPixelObject;

// 3D object structure from lines
  TLineObject = record
    Structure : array of record
      Vertices  : TVectorArray3f;
      Color     : TColor;
      Alpha     : single;
      LineWidth : single;
      LineStyle : TLineStyle;
      Closed    : boolean;
    end;
    Position : TVector3f;
    Rotation : TVector3f;
    Scale    : TVector3f;
    Visible  : boolean;
  end;
  TLineObjectArray = array of TLineObject;

// 3D object structure from triangles
  TTrianglesObject = record
    Structure : array of record
      Vertices : TVectorArray3f;
      Normals  : TVectorArray3f;
      Color    : TColor;
      Alpha    : single;
    end;
    Position    : TVector3f;
    Rotation    : TVector3f;
    Scale       : TVector3f;
    Visible     : boolean;
    IgnoreLight : boolean;
  end;
  TTrianglesObjectArray = array of TTrianglesObject;

// 3D object structure from triangle strips
  TTriangleStripObject = record
    Structure : array of record
      Vertices : TVectorArray3f;
      Normals  : TVectorArray3f;
      Color    : TColor;
      Alpha    : single;
    end;
    Position    : TVector3f;
    Rotation    : TVector3f;
    Scale       : TVector3f;
    Visible     : boolean;
    IgnoreLight : boolean;
  end;
  TTriangleStripObjectArray = array of TTriangleStripObject;

// 3D object structure from triangle fans
  TTriangleFanObject = record
    Structure : array of record
      Vertices : TVectorArray3f;
      Normals  : TVectorArray3f;
      Color    : TColor;
      Alpha    : single;
    end;
    Position    : TVector3f;
    Rotation    : TVector3f;
    Scale       : TVector3f;
    Visible     : boolean;
    IgnoreLight : boolean;
  end;
  TTriangleFanObjectArray = array of TTriangleFanObject;

// 3D object structure from quads
  TQuadsObject = record
    Structure : array of record
      Vertices : TVectorArray3f;
      Normals  : TVectorArray3f;
      Color    : TColor;
      Alpha    : single;
    end;
    Position    : TVector3f;
    Rotation    : TVector3f;
    Scale       : TVector3f;
    Visible     : boolean;
    IgnoreLight : boolean;
  end;
  TQuadsObjectArray = array of TQuadsObject;

// 3D object structure from quad strips
  TQuadStripObject = record
    Structure : array of record
      Vertices : TVectorArray3f;
      Normals  : TVectorArray3f;
      Color    : TColor;
      Alpha    : single;
    end;
    Position    : TVector3f;
    Rotation    : TVector3f;
    Scale       : TVector3f;
    Visible     : boolean;
    IgnoreLight : boolean;
  end;
  TQuadStripObjectArray = array of TQuadStripObject;

// 3D object structure from polygons
  TPolygonObject = record
    Structure : array of record
      Vertices : TVectorArray3f;
      Normals  : TVectorArray3f;
      Color    : TColor;
      Alpha    : single;
    end;
    Position    : TVector3f;
    Rotation    : TVector3f;
    Scale       : TVector3f;
    Visible     : boolean;
    IgnoreLight : boolean;
  end;
  TPolygonObjectArray = array of TPolygonObject;

// Depth Test structure
  TDepthTest = record
    Enable : boolean;
    Func   : cardinal;
  end;

// Blend structure
  TBlend = record
    Enable   : boolean;
    FuncSrc  : cardinal;
    FuncDest : cardinal;
  end;

// AntiAliasing structure
  TAntiAliasing = record
    Enable : boolean;
  end;

// FaceCulling structure
  TFaceCulling = record
    Enable : boolean;
    Func   : cardinal;
  end;

// Fog structure
  TFog = record
    Enable   : boolean;
    Mode     : cardinal;
    Density  : single;
    FogStart : single;
    FogEnd   : single;
    FogIndex : single;
    Color    : TColor;
    CoordSrc : cardinal;
  end;

// Lights structure
  TLights = record
    Enable        : boolean;
    Shading       : cardinal;
    GlobalAmbient : array[0..3] of single;

    // all 8 lights structure
    Light : array[0..7] of record
      Enable       : boolean;
      Func         : cardinal;
      Position     : TVector3f;
      Direction    : TVector3f;
      Ambient      : array[0..3] of single;
      Diffuse      : array[0..3] of single;
      Specular     : array[0..3] of single;
      SpotCutOff   : cardinal;
      SpotExponent : cardinal;
      Const_Atten  : cardinal;
      Linea_Atten  : cardinal;
      Quadr_Atten  : cardinal;
    end;
  end;

type
  TMyOpenGL = Class
  private
    RC : HGLRC;                                // Rendering Context
    DC : HDC;                                  // Device Context

    CallBack : TCallBack;                      // callback for cyclic rendering (controled by inner timer)
    Timer    : TTimer;                         // inner timer for cyclic rendering
    SW       : TStopWatch;                     // rendering deltatime counter

    Canvas       : TControl;                   // drawing output object
    CanvasWidth  : integer;                    // drawing output width
    CanvasHeight : integer;                    // drawing output height

    // rendering
    Rendering      : boolean;                  // process rendering
    RenderFPS      : integer;                  // FPS (0 = maual, 1-xxxx = cyclic rendering)
    RenderNearDist : double;                   // distance rendering from
    RenderFarDist  : double;                   // distance rendering to

    BackgroundColor : TColor;                  // background color

    DeltaTime : double;                        // delta time of 1 rendering cycle

    // camera
    CameraFOV : double;                        // camera field of view

    CameraPosX  : double;                      // camera position X
    CameraPosY  : double;                      // camera position Y
    CameraPosZ  : double;                      // camera position Z
    CameraViewX : double;                      // camera direction vector X
    CameraViewY : double;                      // camera direction vector Y
    CameraViewZ : double;                      // camera direction vector Z
    CameraUpX   : double;                      // camera up vector X
    CameraUpY   : double;                      // camera up vector Y
    CameraUpZ   : double;                      // camera up vector Z
    CameraLeftX : double;                      // camera left vector X
    CameraLeftY : double;                      // camera left vector Y
    CameraLeftZ : double;                      // camera left vector Z

    Camera2DLayerActive : boolean;             // camera ortho 2D layer

    // transformation 3D scene
    TranslateX : double;                       // scene position X
    TranslateY : double;                       // scene position Y
    TranslateZ : double;                       // scene position Z
    ScaleX     : double;                       // scene scale X
    ScaleY     : double;                       // scene scale Y
    ScaleZ     : double;                       // scene scale Z
    RotateX    : double;                       // scene rotation X
    RotateY    : double;                       // scene rotation Y
    RotateZ    : double;                       // scene rotation Z

    // all matrixes
    ProjectionMatrix : TGLMatrixd4;            // projection matrix
    ModelViewMatrix  : TGLMatrixd4;            // model-view matrix
    ViewPortMatrix   : TGLVectori4;            // viewport matrix

    // texture buffer
    TextureBuffer : array of GLuint;           // texture id container

    {Recreate OpenGL context}
    procedure ReCreate;

    {Apply of all transformations}
    procedure ApplyTransformation;

    {OnTimer procedure for inner timer (used for cyclic rendering)}
    procedure TimerTick(Sender: TObject);

  public
    // OpenGL functions
    DepthTest    : TDepthTest;                 // depth buffer settings
    Blend        : TBlend;                     // blend settings
    AntiAliasing : TAntiAliasing;              // antialiasing and rendering quality settings
    FaceCulling  : TFaceCulling;               // face culling settings
    Fog          : TFog;                       // fog settings
    Lights       : TLights;                    // lights settings

    {Creating / destroying MyOpenGL context
       ACallBack is for cyclic rendering (controled by inner timer)}
    constructor Create(const AControl : TControl; const ACallBack : TCallBack);
    destructor  Destroy; override;

    {Get MyOpenGL objects}
    function  GetHDC : HDC;
    function  GetCanvas : TControl;

    {Get actual rendering state
       sets by RenderingBegin and RenderingEnd procedures}
    function  GetRenderingState : boolean;

    {Sets / Gets FPS for cyclic rendering (controled by inner timer)
       value 0 is for manual rendering, 1 to 100 is for cyclic rendering}
    procedure SetRenderFPS(const value : integer);
    function  GetRenderFPS : integer;

    {Scene rendering}
    procedure Render;

    {Sets / Gets minimal and maximal rendering distance}
    procedure SetRenderNearDist(const value : double);
    procedure SetRenderFarDist(const value : double);
    function  GetRenderNearDist : double;
    function  GetRenderFarDist : double;

    {Sets / Gets background color}
    procedure SetBackgroundColor(const color : TColor);
    function  GetBackgroundColor : TColor;

    {Gets delta time of rendering per second}
    function  GetDeltaTime : double;

    {Sets / Gets cameras field of view}
    procedure SetCameraFOV(const value : double);
    function  GetCameraFOV : double;

    {Sets / Gets cameras position and direction vectors}
    procedure SetCameraPosition(const x, y, z : double); overload;
    procedure SetCameraPosition(const v : TVector3f); overload;
    procedure GetCameraPosition(var x, y, z : double); overload;
    function  GetCameraPosition : TVector3f; overload;

    procedure SetCameraViewUpVectors(const viewX, viewY, viewZ, upX, upY, upZ : double); overload;
    procedure SetCameraViewUpVectors(const viewV, upV : TVector3f); overload;
    procedure SetCameraViewLeftVectors(const viewX, viewY, viewZ, leftX, leftY, leftZ : double); overload;
    procedure SetCameraViewLeftVectors(const viewV, leftV : TVector3f); overload;
    procedure SetCameraUpLeftVector(const upX, upY, upZ, leftX, leftY, leftZ : double); overload;
    procedure SetCameraUpLeftVector(const upV, leftV : TVector3f); overload;

    procedure GetCameraViewVector(var x, y, z : double); overload;
    function  GetCameraViewVector : TVector3f; overload;
    procedure GetCameraUpVector(var x, y, z : double); overload;
    function  GetCameraUpVector : TVector3f; overload;
    procedure GetCameraLeftVector(var x, y, z : double); overload;
    function  GetCameraLeftVector : TVector3f; overload;

    {Sets / Gets basic scene transmormation}
    procedure SetTranslate(const x, y, z : double); overload;
    procedure SetTranslate(const v : TVector3f); overload;
    procedure SetScale(const x, y, z : double); overload;
    procedure SetScale(const v : TVector3f); overload;
    procedure SetRotate(const x, y, z : double); overload;
    procedure SetRotate(const v : TVector3f); overload;
    procedure GetTranslate(var x, y, z : double); overload;
    procedure GetTranslate(var v : TVector3f); overload;
    procedure GetScale(var x, y, z : double); overload;
    procedure GetScale(var v : TVector3f); overload;
    procedure GetRotate(var x, y, z : double); overload;
    procedure GetRotate(var v : TVector3f); overload;

    {Camera movement}
    procedure MoveUp(const dist : double);
    procedure MoveDown(const dist : double);
    procedure MoveLeft(const dist : double);
    procedure MoveRight(const dist : double);
    procedure MoveFront(const dist : double);
    procedure MoveBack(const dist : double);

    {Camera rotation
       disabled forceUpVector = rotation like in space
       enabled forceUpVector = rotation like human with freezed up vector defined by fUPx, fUPy, fUPz}
    procedure RotateUp(const angDeg : double);
    procedure RotateDown(const angDeg : double);
    procedure RotateLeft(const angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
    procedure RotateRight(const angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
    procedure RotateClockWise(const angDeg : double);
    procedure RotateCounterClockWise(const angDeg : double);

    {Camera orbit around point in defined distance
       disabled forceUpVector = orbit around like in space
       enabled forceUpVector = orbit around like human with freezed up vector defined by fUPx, fUPy, fUPz}
    procedure OrbitUp(const dist, angDeg : double);
    procedure OrbitDown(const dist, angDeg : double);
    procedure OrbitLeft(const dist, angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
    procedure OrbitRight(const dist, angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);

    {Camera LookAt defined coordinates
       disabled forceUpVector = look at position like in space
       enabled forceUpVector = look at position like human with freezed up vector defined by fUPx, fUPy, fUPz}
    procedure LookAt(const x, y, z : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0); overload;
    procedure LookAt(const v : TVector3f; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0); overload;

    {Calculation from 3D scene to 2D screen layer}
    function  WorldToScreen(const x, y, z : double; var nX, nY : integer) : boolean; overload;
    function  WorldToScreen(const v : TVector3f; var nX, nY : integer) : boolean; overload;

    {Begin and end of rendering, switch to orto 2D camera layer
       every drawings must start with RenderingBegin, after we render all 3D models / objects,
       we can use 2D ortho camera layer with RenderingSetOrthoCameraLayer, after all drawings,
       we must end it with RenderingEnd}
    procedure RenderingBegin;
    procedure RenderingToOrthoCameraLayer;
    procedure RenderingEnd;

    {Settings another OpenGl function}
    procedure SetGLColor(const color : TColor; const alpha : single);
    procedure SetGLPixelSize(const size : single);
    procedure SetGLLineWidth(const width : single);

    {Setting line styles}
    procedure SetGLLineStyle(const factor : Integer; const pattern : word); overload;
    procedure SetGLLineStyle(const style : TLineStyle; const enlarge : Byte = 2); overload;
    procedure SetGLLineStyleEnd;

    {Begin / End of drawing pixels}
    procedure BeginPixels;
    procedure EndPixels;

    {Begin / End of drawing lines
       BeginLines - every two vectors represent one line
       BeginLineStrip - continous line from first to last vector
       BeginLineLoop - continous line from first to last vector with loop to first vector}
    procedure BeginLines;
    procedure BeginLineStrip;
    procedure BeginLineLoop;
    procedure EndLines;

    {Begin / End of drawing triangles
       BeginTriangles - one three vectors triangle
       BeginTriangleStrip - strip from triangles
       BeginTriangleFan - circle from triangles (with shared verte)}
    procedure BeginTriangles;
    procedure BeginTriangleStrip;
    procedure BeginTriangleFan;
    procedure EndTriangles;

    {Begin / End of drawing quads
       BeginQuads - one four vectors quad
       BeginQuadStrip - strip from quads}
    procedure BeginQuads;
    procedure BeginQuadStrip;
    procedure EndQuads;

    {Begin / End of drawing polygons}
    procedure BeginPolygon;
    procedure EndPolygon;

    {Drawings vectors}
    procedure DrawVertex3f(const x, y, z : double); overload;
    procedure DrawVertex3f(const v : TVector3f); overload;
    procedure DrawVertexArray3f(const p : TVectorArray3f);

    {Drawings normals}
    procedure DrawNormal3f(const x, y, z : double); overload;
    procedure DrawNormal3f(const v : TVector3f); overload;
    procedure DrawNormalArray3f(const p : TVectorArray3f);

    {Drawings various objects from predefined structures}
    procedure DrawObject(const obj : TPixelObject); overload;
    procedure DrawObject(const obj : TLineObject); overload;
    procedure DrawObject(const obj : TTrianglesObject; const showNormals : boolean = false; const NormalLen : single = 1); overload;
    procedure DrawObject(const obj : TTriangleStripObject; const showNormals : boolean = false; const NormalLen : single = 1); overload;
    procedure DrawObject(const obj : TTriangleFanObject; const showNormals : boolean = false; const NormalLen : single = 1); overload;
    procedure DrawObject(const obj : TQuadsObject; const showNormals : boolean = false; const NormalLen : single = 1); overload;
    procedure DrawObject(const obj : TQuadStripObject; const showNormals : boolean = false; const NormalLen : single = 1); overload;
    procedure DrawObject(const obj : TPolygonObject; const showNormals : boolean = false; const NormalLen : single = 1); overload;
    procedure DrawObjectArray(const objArray : TPixelObjectArray); overload;
    procedure DrawObjectArray(const objArray : TLineObjectArray); overload;
    procedure DrawObjectArray(const objArray : TTrianglesObjectArray); overload;
    procedure DrawObjectArray(const objArray : TTriangleStripObjectArray); overload;
    procedure DrawObjectArray(const objArray : TTriangleFanObjectArray); overload;
    procedure DrawObjectArray(const objArray : TQuadsObjectArray); overload;
    procedure DrawObjectArray(const objArray : TQuadStripObjectArray); overload;
    procedure DrawObjectArray(const objArray : TPolygonObjectArray); overload;

    {Return vector in camera front direction in defined distance Dist}
    procedure GetLookAtDistVector(const Dist : double; var x, y, z : double); overload;
    function  GetLookAtDistVector(const Dist : double) : TVector3f; overload;

    {Vector position test relative to camera view, return +X, 0, -X depends on function}
    function  VectorUpperLowerSide(const x, y, z : double) : double; overload;
    function  VectorUpperLowerSide(const v : TVector3f) : double; overload;
    function  VectorLeftRightSide(const x, y, z : double) : double; overload;
    function  VectorLeftRightSide(const v : TVector3f) : double; overload;
    function  VectorFrontBehindSide(const x, y, z : double) : double; overload;
    function  VectorFrontBehindSide(const v : TVector3f) : double; overload;

    {Calculate mouse XY coordinates and return vector and direction as simple mouse picking}
    procedure GetMouseRay(const mouseX, mouseY : integer; var vStart, vEnd : TVector3f);
  end;

{Create various line objects}
procedure CreateLineObject_Axes(const length : double; var obj : TLineObject);
procedure CreateLineObject_Grid(const step : double; const blocks : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Line(const x1,y1,z1,x2,y2,z2 : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Arc(const radius, angleStart, angleSweep : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Circle(const radius : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Cross(const size : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Cube(const sizeX, sizeY, sizeZ : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Cylinder(const radius, height : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Pie(const radius : double; const segments : integer; const CutoutPercents : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Ring(const radiusInner, radiusOuter : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
procedure CreateLineObject_Star(const radiusInner, radiusOuter : double; const tips : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);

{Create various 3D objects}
procedure Create3DObject_Plane(const height, width : double; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TQuadsObject);
procedure Create3DObject_Cube(const sizeX, sizeY, sizeZ : double; const pFrontColor, pBackColor, pTopColor, pBottomColor, pLeftColor, pRightColor : TColor; const pAlpha : single; var obj : TQuadsObject);
procedure Create3DObject_Cylinder(const radius, height : double; const segments : integer; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TQuadStripObject);
procedure Create3DObject_Pie(const radius : double; const segments : integer; const CutoutPercents : double; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TTriangleFanObject);
procedure Create3DObject_Ring(const radiusInner, radiusOuter : double; const segments : integer; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TQuadStripObject);
procedure Create3DObject_Star(const radiusInner, radiusOuter : double; const tips : integer; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TPolygonObject);

{Create base xyz vector}
procedure MakeVector3f(const x, y, z : double; var v : TVector3f); overload;
function  MakeVector3f(const x, y, z : double) : TVector3f; overload;

{Convecrtion between OpenGL and Win color}
function  GLColorToTColor(const r, g, b : single) : TColor;
procedure TColorToGLColor(const color : TColor; const alpha : single; var r, g, b, w : single);

{Compare two vectors}
function  VectorEquals(const x1, y1, z1, x2, y2, z2 : double; const epsilon : double = 0.00001) : boolean; overload;
function  VectorEquals(const v1, v2 : TVector3f; const epsilon : double = 0.00001) : boolean; overload;

{Distance between two vectors}
function  VectorDistance(const x1, y1, z1, x2, y2, z2 : double) : double; overload;
function  VectorDistance(const v1, v2 : TVector3f) : double; overload;

{Return normalized vector}
procedure VectorNormalize(const x, y, z : double; var vX, vY, vZ : double); overload;
procedure VectorNormalize(const v1 : TVector3f; var v2 : TVector3f); overload;
function  VectorNormalize(const v : TVector3f) : TVector3f; overload;

{Summary of two vectors V1 + V2}
procedure VectorAdd(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var nX, nY, nZ : double); overload;
procedure VectorAdd(const v1, v2 : TVector3f; var v3 : TVector3f); overload;
function  VectorAdd(const v1, v2 : TVector3f) : TVector3f; overload;

{Subtraction of two vectors V1 - V2}
procedure VectorSubtract(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var nX, nY, nZ : double); overload;
procedure VectorSubtract(const v1, v2 : TVector3f; var v3 : TVector3f); overload;
function  VectorSubtract(const v1, v2 : TVector3f) : TVector3f; overload;

{Vector negation}
procedure VectorNegate(const vX, vY, vZ : double; var nX, nY, nZ : double); overload;
procedure VectorNegate(const v1 : TVector3f; var v2 : TVector3f); overload;
function  VectorNegate(const v : TVector3f) : TVector3f; overload;

{Calculate vector cross product}
procedure VectorCrossProduct(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var nX, nY, nZ : double); overload;
procedure VectorCrossProduct(const v1, v2 : TVector3f; var v3 : TVector3f); overload;
function  VectorCrossProduct(const v1, v2 : TVector3f) : TVector3f; overload;

{Point projection from Src vector in Dst direction in length Dist}
procedure ProjectPoint(const Srcx, Srcy, Srcz, Dstx, Dsty, Dstz, Dist : double; out Nx, Ny, Nz : double); overload;
procedure ProjectPoint(const Src, Dst : TVector3f; const Dist : double; out vOut : TVector3f); overload;

{Vector xyz rotation in defined angles for each axis}
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const x, y, z : double; out Nx, Ny, Nz : double); overload;
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const vIn : TVector3f; out vOut : TVector3f); overload;

{Vector xyz rotation in defined angles for each axis around vector O}
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const x, y, z, Ox, Oy, Oz : double; out Nx, Ny, Nz : double); overload;
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const vIn, vAround : TVector3f; out vOut : TVector3f); overload;

{Vector n rotation around vector axis in defined angle}
procedure RotateVector(var nX, nY, nZ : double; const axisX, axisY, axisZ, angleRad : double); overload;
procedure RotateVector(var vIn : TVector3f; const vAxis : TVector3f; const angleRad : double); overload;

{Linear interpolation of float value from start to stop, delta t is between 0 and 1}
function  Lerp(const start, stop, t : double) : double;

{Linear interpolation of angle from start to stop, delta t is between 0 and 1}
function  AngleLerp(startRad, stopRad : double; const t : double) : double;

{Return cartesian quadrant (1 to 4) of defined angle in degrees}
function  Quadrant(const angle : double) : integer;

{Return angle in degrees of XY point in cartesian coordinate system}
function  CartesianAngle(const x, y : double) : double;

{Return XY point lying in defined angle in degrees and distance from zero in cartesian coordinate system}
procedure CartesianPoint(const angle, radius : double; var x, y : double);

{Normalizing radian angle in range -PI to +PI}
function  NormalizeAngle(const angleRad : double) : double;

{Linear interpolation between two vectors, delta t is between 0 and 1}
procedure VectorLerp(const v1X, v1Y, v1Z, v2X, v2Y, v2Z, t : double; var vrX, vrY, vrZ : double); overload;
procedure VectorLerp(const v1, v2 : TVector3f; const t : double; var vOut : TVector3f); overload;
function  VectorLerp(const v1, v2 : TVector3f; const t : double) : TVector3f; overload;

{Return angle between vectors 1 and 3, vector 2 is a corner (result angle is in degrees)}
function  AngleBetweenTwoVectors(x1, y1, z1, x2, y2, z2, x3, y3, z3 : double) : double; overload;
function  AngleBetweenTwoVectors(v1, v2, v3 : TVector3f) : double; overload;

{Create rotation matrix from defined vector and angle}
procedure CreateRotationMatrix(const axisX, axisY, axisZ, angleRad : double; var mX_X, mX_Y, mX_Z, mX_W, mY_X, mY_Y, mY_Z, mY_W, mZ_X, mZ_Y, mZ_Z, mZ_W, mW_X, mW_Y, mW_Z, mW_W : double); overload;
procedure CreateRotationMatrix(const axisVec : TVector3f; const angleRad : double; var matrixOut : TMatrix16f); overload;

{Transform homogeneous vector by multiplication with matrix}
procedure VectorMatrixTransform(const vX, vY, vZ, mX_X, mX_Y, mX_Z, mX_W, mY_X, mY_Y, mY_Z, mY_W, mZ_X, mZ_Y, mZ_Z, mZ_W, mW_X, mW_Y, mW_Z, mW_W : double; var nX, nY, nZ : double); overload;
procedure VectorMatrixTransform(const Vec : TVector3f; const Mat : TMatrix16f; var vOut : TVector3f); overload;

{Dot product of two vectors}
function  VectorDotProduct(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double) : double; overload;
procedure VectorDotProduct(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var dot : double); overload;
function  VectorDotProduct(const v1, v2 : TVector3f) : double; overload;
procedure VectorDotProduct(const v1, v2 : TVector3f; var dot : double); overload;

{Direction vector (normalized) from two vectors}
procedure DirectionVector(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var vrX, vrY, vrZ : double); overload;
procedure DirectionVector(const v1, v2 : TVector3f; var vOut : TVector3f); overload;
function  DirectionVector(const v1, v2 : TVector3f) : TVector3f; overload;

{Application position, rotation and scale to defined vector
   rotation can be aplied at once or step by step to achieve the desired result}
function  VectorCombinePosRotSca(const v, position, rotation, scale : TVector3f) : TVector3f; overload;
function  VectorCombinePosRotSca(const v, position, rotationX, rotationY, rotationZ, scale : TVector3f) : TVector3f; overload;

{Vector position from plane defined by two vectors
   result 0 = vector lays on same plane
   result +1 = vector lays on first side from plane
   result -1 = vector lays on second side from plane}
function  VectorSide_3D(const x1, y1, z1, x2, y2, z2, x3, y3, z3, Px, Py, Pz : double; const epsilon : double = 0.00001) : integer; overload;
function  VectorSide_3D(const v1, v2, v3, vP : TVector3f; const epsilon : double = 0.00001) : integer; overload;

{Colinear test between line defined by vectors v1 and v2 and point defined by vector v3}
function  CollinearTest(const x1, y1, z1, x2, y2, z2, x3, y3, z3 : double; const epsilon : double = 0.00001) : boolean; overload;
function  CollinearTest(const v1, v2, v3 : TVector3f; const epsilon : double = 0.00001) : boolean; overload;

{Coplanar test between plane defined by first three vectors and other vectors in array}
function  CoplanarTest(const vArr : TVectorArray3f; const epsilon : double = 0.00001) : boolean; overload;
function  CoplanarTest(const vPlane, vArr : TVectorArray3f; const epsilon : double = 0.00001) : boolean; overload;

{Test if vector exists in vector array and result its position index in array}
function  VectorInVectorArrayExists(const v : TVector3f; const Arr : TVectorArray3f; var arr_index : integer; const epsilon : double = 0.00001) : boolean;

{Test if vector exists in array of vector array and result its position index in both arrays}
function  VectorInVectorArrayArrayExists(const v : TVector3f; const Arr : TVectorArrayArray3f; var first_arr_index, second_arr_index : integer; const epsilon : double = 0.00001) : boolean;

implementation

{----- TMyOpenGL.ReCreate -----}
procedure TMyOpenGL.ReCreate;
begin
  // new Device Context
  DC := GetDC(TWinControl(Canvas).Handle);

  // Canvas dimensions
  CanvasWidth  := Canvas.ClientWidth;
  CanvasHeight := Canvas.ClientHeight;

  // dglOpenGL - init
  if not InitOpenGL then raise Exception.Create('Error occured while OpenGL initialization!');

  // new Rendering Context
  RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0);
  ActivateRenderingContext(DC, RC);

  // viewport setting
  glViewPort(0, 0, CanvasWidth, CanvasHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(CameraFOV, CanvasWidth / CanvasHeight, RenderNearDist, RenderFarDist);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
end;

{----- TMyOpenGL.ApplyTransformation -----}
procedure TMyOpenGL.ApplyTransformation;
var
  a : integer;
  i : integer;
  elapsed : int64;
begin
  if not Rendering then Exit;

  // deltatime setting
  if RenderFPS = 0 then DeltaTime := 1 // for manual rendering
  else
  begin
    // for cyclic rendering
    elapsed := SW.ElapsedMilliseconds;
    if elapsed = 0 then DeltaTime := 1
                   else DeltaTime := (1000 / RenderFPS) / elapsed;
    SW.Reset;
    SW.Start;
  end;

  // set identity matrix
  glLoadIdentity;

  // correction + normalization cameras direction vectors
  VectorNormalize(CameraViewX, CameraViewY, CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);
  VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraUpX, CameraUpY, CameraUpZ, CameraLeftX, CameraLeftY, CameraLeftZ);
  VectorNormalize(-CameraLeftX, -CameraLeftY, -CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);
  VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraLeftX, CameraLeftY, CameraLeftZ, CameraUpX, CameraUpY, CameraUpZ);
  VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);

	// camera setting
	gluLookAt(CameraPosX, CameraPosY, CameraPosZ, CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ, CameraUpX, CameraUpY, CameraUpZ);

  // scene transformation
  glRotatef(RotateZ, 0.0, 0.0, 1.0);
  glRotatef(RotateY, 0.0, 1.0, 0.0);
  glRotatef(RotateX, 1.0, 0.0, 0.0);
  glTranslatef(TranslateX, TranslateY, TranslateZ);
  glScalef(ScaleX, ScaleY, ScaleZ);

  // store all matrixes for later use (if needed)
  for a := 0 to 3 do
  begin
    for i := 0 to 3 do
    begin
      ProjectionMatrix[a, i] := 1;
      ModelViewMatrix[a, i]  := 1;
    end;
    ViewPortMatrix[a] := 1;
  end;

  // setting all matrixes
  glGetDoublev(GL_PROJECTION_MATRIX, @ProjectionMatrix);
  glGetDoublev(GL_MODELVIEW_MATRIX,  @ModelViewMatrix);
  glGetIntegerv(GL_VIEWPORT,         @ViewPortMatrix);
end;

{----- TMyOpenGL.TimerTick -----}
procedure TMyOpenGL.TimerTick(Sender: TObject);
begin
  Render;
end;

{----- TMyOpenGL.Create -----}
constructor TMyOpenGL.Create(const AControl : TControl; const ACallBack : TCallBack);
var
  i : integer;
begin
  // AControl must be TWinControl
  Canvas := AControl;
  if not (Canvas is TWinControl) then raise Exception.Create('OpenGL Canvas must be TWinControl type!');

  // callback procedure for cyclic rendering (controled by inner timer)
  CallBack := ACallBack;

  // creating inner timer
  Timer := TTimer.Create(nil);
  with Timer do
  begin
    Enabled  := false;
    Interval := 0;
    OnTimer  := TimerTick;
  end;

  // default settings

  CanvasWidth  := -1;
  CanvasHeight := -1;

  Rendering      := false;
  RenderFPS      := 0;
  RenderNearDist := 1;
  RenderFarDist  := 10000;

  BackgroundColor := clBlack;

  CameraFOV := 45;

  CameraPosX  := 0;
  CameraPosY  := 0;
  CameraPosZ  := 0;
  CameraViewX := 0;
  CameraViewY := 0;
  CameraViewZ := 1;
  CameraUpX   := 0;
  CameraUpY   := 1;
  CameraUpZ   := 0;
  CameraLeftX := 1;
  CameraLeftY := 0;
  CameraLeftZ := 0;

  TranslateX := 0;
  TranslateY := 0;
  TranslateZ := 0;
  ScaleX     := 1;
  ScaleY     := 1;
  ScaleZ     := 1;
  RotateX    := 0;
  RotateY    := 0;
  RotateZ    := 0;

  // first OpenGL initialisation after Create
  ReCreate;

  // default OpenGL settings
  glPushAttrib(GL_ENABLE_BIT);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glDisable(GL_BLEND);
  glDisable(GL_FOG);
  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_TEXTURE_1D);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_TEXTURE_3D);
  glDisable(GL_TEXTURE_CUBE_MAP_ARB);
  glDisable(GL_SMOOTH);
  glDisable(GL_LINE_SMOOTH);
  glDisable(GL_POINT_SMOOTH);
  glDisable(GL_POLYGON_SMOOTH);
  glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

  // depth test setting
  DepthTest.Enable := true;
  DepthTest.Func   := GL_LEQUAL;

  // blend setting
  Blend.Enable   := false;
  Blend.FuncSrc  := GL_SRC_ALPHA;
  Blend.FuncDest := GL_ONE_MINUS_SRC_ALPHA;

  // antialiasing setting
  AntiAliasing.Enable := false;

  // face culling setting
  FaceCulling.Enable := true;
  FaceCulling.Func   := GL_BACK;

  // fog setting
  Fog.Enable   := false;
  Fog.Mode     := GL_EXP;
  Fog.Density  := 0.01;
  Fog.FogStart := 500;
  Fog.FogEnd   := 1000;
  Fog.FogIndex := 0;
  Fog.Color    := clBlack;
  Fog.CoordSrc := GL_FOG_COORD_SRC;

  // lights setting
  Lights.Enable           := false;
  Lights.Shading          := GL_SMOOTH;
  Lights.GlobalAmbient[0] := 0.2;
  Lights.GlobalAmbient[1] := 0.2;
  Lights.GlobalAmbient[2] := 0.2;
  Lights.GlobalAmbient[3] := 1.0;

  // creating all 8 lights with default setting
  for i := 0 to 7 do
    with Lights.Light[i] do
    begin
      Enable := false;
      case i of
        0 : Func := GL_LIGHT0;
        1 : Func := GL_LIGHT1;
        2 : Func := GL_LIGHT2;
        3 : Func := GL_LIGHT3;
        4 : Func := GL_LIGHT4;
        5 : Func := GL_LIGHT5;
        6 : Func := GL_LIGHT6;
        7 : Func := GL_LIGHT7;
      end;
      Position.X   := 0;
      Position.Y   := 0;
      Position.Z   := 0;
      Direction.X  := 0;
      Direction.Y  := 0;
      Direction.Z  := -1;
      Ambient[0]   := 0; Ambient[1]  := 0; Ambient[2]  := 0; Ambient[3]  := 1;
      Diffuse[0]   := 1; Diffuse[1]  := 1; Diffuse[2]  := 1; Diffuse[3]  := 1;
      Specular[0]  := 1; Specular[1] := 1; Specular[2] := 1; Specular[3] := 1;
      SpotCutOff   := 180;
      SpotExponent := 0;
      Const_Atten  := 1;
      Linea_Atten  := 0;
      Quadr_Atten  := 0;
    end;

  // camera ortho layer deactivation
  Camera2DLayerActive := false;

  // texture buffer
  SetLength(TextureBuffer, 0);
end;

{----- TMyOpenGL.Destroy -----}
destructor TMyOpenGL.Destroy;
begin
  // inner timer
  FreeAndNil(Timer);

  // callback procedure
  CallBack := nil;

  // OpenGL context
  DeactivateRenderingContext;
  if RC <> 0 then DestroyRenderingContext(RC);
  if (DC = 1) and Assigned(Canvas) then ReleaseDC(Canvas.Parent.Handle, DC);

  inherited;
end;

{----- TMyOpenGL.GetDC -----}
function TMyOpenGL.GetHDC : HDC;
begin
  result := DC;
end;

{----- TMyOpenGL.GetCanvas -----}
function TMyOpenGL.GetCanvas : TControl;
begin
  result := Canvas;
end;

{----- TMyOpenGL.GetRenderingState -----}
function TMyOpenGL.GetRenderingState : boolean;
begin
  result := Rendering;
end;

{----- TMyOpenGL.SetRenderFPS -----}
procedure TMyOpenGL.SetRenderFPS(const value : integer);
begin
  // min max correction between 0 and 100 FPS
       if value < 0   then RenderFPS := 0
  else if value > 100 then RenderFPS := 100
  else                     RenderFPS := value;

  // activation / deactivation inner timeru
  if RenderFPS = 0 then
  begin
    // manual rendering
    Timer.Enabled := false;
    DeltaTime     := 1;
    SW.Stop;
  end
  else
  begin
    // cyclic rendering
    Timer.Interval := 1000 div RenderFPS; // calculation from FPS to milliseconds
    Timer.Enabled  := true;
    DeltaTime      := 1;
    SW.Start;
  end;
end;

{----- TMyOpenGL.GetRenderFPS -----}
function TMyOpenGL.GetRenderFPS : integer;
begin
  result := RenderFPS;
end;

{----- TMyOpenGL.Render -----}
procedure TMyOpenGL.Render;
begin
  if Rendering then Exit;

  if Assigned(CallBack) then
  begin
    CallBack(DC, Canvas);

    // if manual rendering is set, we call callback twice to force second redraw
    // possibly OpenGL bug (some PCs ignore first rendering but second works)
    if RenderFPS = 0 then
    begin
      Canvas.Update;
      CallBack(DC, Canvas);
    end;
  end;
end;

{----- TMyOpenGL.SetRenderNearDist -----}
procedure TMyOpenGL.SetRenderNearDist(const value : double);
begin
  // min max correction against rendering maximal distance
       if value < 0.1           then RenderNearDist := 0.1
  else if value > RenderFarDist then RenderNearDist := RenderFarDist - 0.1
  else                               RenderNearDist := value;

  // change minimal rendering distance + reset canvas object for new ReCreate
  CanvasWidth  := -1;
  CanvasHeight := -1;
end;

{----- TMyOpenGL.SetRenderFarDist -----}
procedure TMyOpenGL.SetRenderFarDist(const value : double);
begin
  // min max correction against minimal rendering distance
  if value < RenderNearDist then RenderFarDist := RenderNearDist + 0.1
                            else RenderFarDist := value;

  // reset canvas object for new ReCreate
  CanvasWidth  := -1;
  CanvasHeight := -1;
end;

{----- TMyOpenGL.GetRenderNearDist -----}
function TMyOpenGL.GetRenderNearDist : double;
begin
  result := RenderNearDist;
end;

{----- TMyOpenGL.GetRenderFarDist -----}
function TMyOpenGL.GetRenderFarDist : double;
begin
  result := RenderFarDist;
end;

{----- TMyOpenGL.SetBackgroundColor -----}
procedure TMyOpenGL.SetBackgroundColor(const color : TColor);
begin
  BackgroundColor := color;
end;

{----- TMyOpenGL.GetBackgroundColor -----}
function TMyOpenGL.GetBackgroundColor : TColor;
begin
  result := BackgroundColor;
end;

{----- TMyOpenGL.GetDeltaTime -----}
function TMyOpenGL.GetDeltaTime : double;
begin
  result := DeltaTime;
end;

{----- TMyOpenGL.SetCameraFOV -----}
procedure TMyOpenGL.SetCameraFOV(const value : double);
begin
  // min max cerrection
       if value < 1   then CameraFOV := 1
  else if value > 179 then CameraFOV := 179
  else                     CameraFOV := value;

  // reset canvas object for new ReCreate
  CanvasWidth  := -1;
  CanvasHeight := -1;
end;

{----- TMyOpenGL.GetCameraFOV -----}
function TMyOpenGL.GetCameraFOV : double;
begin
  result := CameraFOV;
end;

{----- TMyOpenGL.SetCameraPosition -----}
procedure TMyOpenGL.SetCameraPosition(const x, y, z : double);
begin
  CameraPosX := x;
  CameraPosY := y;
  CameraPosZ := z;
end;

{----- TMyOpenGL.SetCameraPosition -----}
procedure TMyOpenGL.SetCameraPosition(const v : TVector3f);
begin
  CameraPosX := v.X;
  CameraPosY := v.Y;
  CameraPosZ := v.Z;
end;

{----- TMyOpenGL.GetCameraPosition -----}
procedure TMyOpenGL.GetCameraPosition(var x, y, z : double);
begin
  x := CameraPosX;
  y := CameraPosY;
  z := CameraPosZ;
end;

{----- TMyOpenGL.GetCameraPosition -----}
function TMyOpenGL.GetCameraPosition : TVector3f;
begin
  result := MakeVector3f(CameraPosX, CameraPosY, CameraPosZ);
end;

{----- TMyOpenGL.SetCameraViewUpVectors -----}
procedure TMyOpenGL.SetCameraViewUpVectors(const viewX, viewY, viewZ, upX, upY, upZ : double);
begin
  // two direction vectors as input, third one we must to calculate
  VectorNormalize(viewX, viewY, viewZ, CameraViewX, CameraViewY, CameraViewZ);
  VectorNormalize(upX, upY, upZ, CameraUpX, CameraUpY, CameraUpZ);
  VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraUpX, CameraUpY, CameraUpZ, CameraLeftX, CameraLeftY, CameraLeftZ);
  VectorNormalize(CameraLeftX, CameraLeftY, CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);
end;

{----- TMyOpenGL.SetCameraViewUpVectors -----}
procedure TMyOpenGL.SetCameraViewUpVectors(const viewV, upV : TVector3f);
begin
  SetCameraViewUpVectors(viewV.X, viewV.Y, viewV.Z, upV.X, upV.Y, upV.Z);
end;

{----- TMyOpenGL.SetCameraViewLeftVectors -----}
procedure TMyOpenGL.SetCameraViewLeftVectors(const viewX, viewY, viewZ, leftX, leftY, leftZ : double);
begin
  // two direction vectors as input, third one we must to calculate
  VectorNormalize(viewX, viewY, viewZ, CameraViewX, CameraViewY, CameraViewZ);
  VectorNormalize(leftX, leftY, leftZ, CameraLeftX, CameraLeftY, CameraLeftZ);
  VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraLeftX, CameraLeftY, CameraLeftZ, CameraUpX, CameraUpY, CameraUpZ);
  VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);
end;

{----- TMyOpenGL.SetCameraViewLeftVectors -----}
procedure TMyOpenGL.SetCameraViewLeftVectors(const viewV, leftV : TVector3f);
begin
  SetCameraViewLeftVectors(viewV.X, viewV.Y, viewV.Z, leftV.X, leftV.Y, leftV.Z);
end;

{----- TMyOpenGL.SetCameraUpLeftVector -----}
procedure TMyOpenGL.SetCameraUpLeftVector(const upX, upY, upZ, leftX, leftY, leftZ : double);
begin
  // two direction vectors as input, third one we must to calculate
  VectorNormalize(upX, upY, upZ, CameraUpX, CameraUpY, CameraUpZ);
  VectorNormalize(leftX, leftY, leftZ, CameraLeftX, CameraLeftY, CameraLeftZ);
  VectorCrossProduct(CameraUpX, CameraUpY, CameraUpZ, CameraLeftX, CameraLeftY, CameraLeftZ, CameraViewX, CameraViewY, CameraViewZ);
  VectorNormalize(-CameraViewX, -CameraViewY, -CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);
end;

{----- TMyOpenGL.SetCameraUpLeftVector -----}
procedure TMyOpenGL.SetCameraUpLeftVector(const upV, leftV : TVector3f);
begin
  SetCameraUpLeftVector(upV.X, upV.Y, upV.Z, leftV.X, leftV.Y, leftV.Z);
end;

{----- TMyOpenGL.GetCameraViewVector -----}
procedure TMyOpenGL.GetCameraViewVector(var x, y, z : double);
begin
  x := CameraViewX;
  y := CameraViewY;
  z := CameraViewZ;
end;

{----- TMyOpenGL.GetCameraViewVector -----}
function  TMyOpenGL.GetCameraViewVector : TVector3f;
begin
  Result := MakeVector3f(CameraViewX, CameraViewY, CameraViewZ);
end;

{----- TMyOpenGL.GetCameraUpVector -----}
procedure TMyOpenGL.GetCameraUpVector(var x, y, z : double);
begin
  x := CameraUpX;
  y := CameraUpY;
  z := CameraUpZ;
end;

{----- TMyOpenGL.GetCameraUpVector -----}
function  TMyOpenGL.GetCameraUpVector : TVector3f;
begin
  Result := MakeVector3f(CameraUpX, CameraUpY, CameraUpZ);
end;

{----- TMyOpenGL.GetCameraLeftVector -----}
procedure TMyOpenGL.GetCameraLeftVector(var x, y, z : double);
begin
  x := CameraLeftX;
  y := CameraLeftY;
  z := CameraLeftZ;
end;

{----- TMyOpenGL.GetCameraLeftVector -----}
function  TMyOpenGL.GetCameraLeftVector : TVector3f;
begin
  Result := MakeVector3f(CameraLeftX, CameraLeftY, CameraLeftZ);
end;

{----- TMyOpenGL.SetTranslate -----}
procedure TMyOpenGL.SetTranslate(const x, y, z : double);
begin
  TranslateX := x;
  TranslateY := y;
  TranslateZ := z;
end;

{----- TMyOpenGL.SetTranslate -----}
procedure TMyOpenGL.SetTranslate(const v : TVector3f);
begin
  SetTranslate(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.SetScale -----}
procedure TMyOpenGL.SetScale(const x, y, z : double);
begin
  ScaleX := x;
  ScaleY := y;
  ScaleZ := z;
end;

{----- TMyOpenGL.SetScale -----}
procedure TMyOpenGL.SetScale(const v : TVector3f);
begin
  SetScale(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.SetRotate -----}
procedure TMyOpenGL.SetRotate(const x, y, z : double);
begin
  RotateX := x;
  RotateY := y;
  RotateZ := z;
end;

{----- TMyOpenGL.SetRotate -----}
procedure TMyOpenGL.SetRotate(const v : TVector3f);
begin
  SetRotate(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.GetTranslate -----}
procedure TMyOpenGL.GetTranslate(var x, y, z : double);
begin
  x := TranslateX;
  y := TranslateY;
  z := TranslateZ;
end;

{----- TMyOpenGL.GetTranslate -----}
procedure TMyOpenGL.GetTranslate(var v : TVector3f);
begin
  GetTranslate(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.GetScale -----}
procedure TMyOpenGL.GetScale(var x, y, z : double);
begin
  x := ScaleX;
  y := ScaleY;
  z := ScaleZ;
end;

{----- TMyOpenGL.GetScale -----}
procedure TMyOpenGL.GetScale(var v : TVector3f);
begin
  GetScale(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.GetRotate -----}
procedure TMyOpenGL.GetRotate(var x, y, z : double);
begin
  x := RotateX;
  y := RotateY;
  z := RotateZ;
end;

{----- TMyOpenGL.GetRotate -----}
procedure TMyOpenGL.GetRotate(var v : TVector3f);
begin
  GetRotate(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.MoveUp -----}
procedure TMyOpenGL.MoveUp(const dist : double);
begin
  CameraPosX := CameraPosX + (CameraUpX * dist);
  CameraPosY := CameraPosY + (CameraUpY * dist);
  CameraPosZ := CameraPosZ + (CameraUpZ * dist);
end;

{----- TMyOpenGL.MoveDown -----}
procedure TMyOpenGL.MoveDown(const dist : double);
begin
  CameraPosX := CameraPosX + (CameraUpX * -dist);
  CameraPosY := CameraPosY + (CameraUpY * -dist);
  CameraPosZ := CameraPosZ + (CameraUpZ * -dist);
end;

{----- TMyOpenGL.MoveLeft -----}
procedure TMyOpenGL.MoveLeft(const dist : double);
begin
  CameraPosX := CameraPosX + (CameraLeftX * dist);
  CameraPosY := CameraPosY + (CameraLeftY * dist);
  CameraPosZ := CameraPosZ + (CameraLeftZ * dist);
end;

{----- TMyOpenGL.MoveRight -----}
procedure TMyOpenGL.MoveRight(const dist : double);
begin
  CameraPosX := CameraPosX + (CameraLeftX * -dist);
  CameraPosY := CameraPosY + (CameraLeftY * -dist);
  CameraPosZ := CameraPosZ + (CameraLeftZ * -dist);
end;

{----- TMyOpenGL.MoveFront -----}
procedure TMyOpenGL.MoveFront(const dist : double);
begin
  CameraPosX := CameraPosX + (CameraViewX * dist);
  CameraPosY := CameraPosY + (CameraViewY * dist);
  CameraPosZ := CameraPosZ + (CameraViewZ * dist);
end;

{----- TMyOpenGL.MoveBack -----}
procedure TMyOpenGL.MoveBack(const dist : double);
begin
  CameraPosX := CameraPosX + (CameraViewX * -dist);
  CameraPosY := CameraPosY + (CameraViewY * -dist);
  CameraPosZ := CameraPosZ + (CameraViewZ * -dist);
end;

{----- TMyOpenGL.RotateUp -----}
procedure TMyOpenGL.RotateUp(const angDeg : double);
begin
  RotateVector(CameraUpX, CameraUpY, CameraUpZ, CameraLeftX, CameraLeftY, CameraLeftZ, DegToRad(angDeg));
  VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);

  RotateVector(CameraViewX, CameraViewY, CameraViewZ, CameraLeftX, CameraLeftY, CameraLeftZ, DegToRad(angDeg));
  VectorNormalize(CameraViewX, CameraViewY, CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);
end;

{----- TMyOpenGL.RotateDown -----}
procedure TMyOpenGL.RotateDown(const angDeg : double);
begin
  RotateVector(CameraUpX, CameraUpY, CameraUpZ, CameraLeftX, CameraLeftY, CameraLeftZ, DegToRad(-angDeg));
  VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);

  RotateVector(CameraViewX, CameraViewY, CameraViewZ, CameraLeftX, CameraLeftY, CameraLeftZ, DegToRad(-angDeg));
  VectorNormalize(CameraViewX, CameraViewY, CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);
end;

{----- TMyOpenGL.RotateLeft -----}
procedure TMyOpenGL.RotateLeft(const angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
begin
  if not forceUpVector then
  begin
    RotateVector(CameraLeftX, CameraLeftY, CameraLeftZ, CameraUpX, CameraUpY, CameraUpZ, DegToRad(-angDeg));
    VectorNormalize(CameraLeftX, CameraLeftY, CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);

    RotateVector(CameraViewX, CameraViewY, CameraViewZ, CameraUpX, CameraUpY, CameraUpZ, DegToRad(-angDeg));
    VectorNormalize(CameraViewX, CameraViewY, CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);
  end
  else
  begin
    RotateVector(CameraLeftX, CameraLeftY, CameraLeftZ, fUpX, fUpY, fUpZ, DegToRad(-angDeg));
    VectorNormalize(CameraLeftX, CameraLeftY, CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);

    RotateVector(CameraViewX, CameraViewY, CameraViewZ, fUpX, fUpY, fUpZ, DegToRad(-angDeg));
    VectorNormalize(CameraViewX, CameraViewY, CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);

    VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraLeftX, CameraLeftY, CameraLeftZ, CameraUpX, CameraUpY, CameraUpZ);
    VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);
  end;
end;

{----- TMyOpenGL.RotateRight -----}
procedure TMyOpenGL.RotateRight(const angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
begin
  if not forceUpVector then
  begin
    RotateVector(CameraLeftX, CameraLeftY, CameraLeftZ, CameraUpX, CameraUpY, CameraUpZ, DegToRad(angDeg));
    VectorNormalize(CameraLeftX, CameraLeftY, CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);

    RotateVector(CameraViewX, CameraViewY, CameraViewZ, CameraUpX, CameraUpY, CameraUpZ, DegToRad(angDeg));
    VectorNormalize(CameraViewX, CameraViewY, CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);
  end
  else
  begin
    RotateVector(CameraLeftX, CameraLeftY, CameraLeftZ, fUpX, fUpY, fUpZ, DegToRad(angDeg));
    VectorNormalize(CameraLeftX, CameraLeftY, CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);

    RotateVector(CameraViewX, CameraViewY, CameraViewZ, fUpX, fUpY, fUpZ, DegToRad(angDeg));
    VectorNormalize(CameraViewX, CameraViewY, CameraViewZ, CameraViewX, CameraViewY, CameraViewZ);

    VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraLeftX, CameraLeftY, CameraLeftZ, CameraUpX, CameraUpY, CameraUpZ);
    VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);
  end;
end;

{----- TMyOpenGL.RotateClockWise -----}
procedure TMyOpenGL.RotateClockWise(const angDeg : double);
begin
  RotateVector(CameraLeftX, CameraLeftY, CameraLeftZ, CameraViewX, CameraViewY, CameraViewZ, DegToRad(-angDeg));
  VectorNormalize(CameraLeftX, CameraLeftY, CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);

  RotateVector(CameraUpX, CameraUpY, CameraUpZ, CameraViewX, CameraViewY, CameraViewZ, DegToRad(-angDeg));
  VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);
end;

{----- TMyOpenGL.RotateCounterClockWise -----}
procedure TMyOpenGL.RotateCounterClockWise(const angDeg : double);
begin
  RotateVector(CameraLeftX, CameraLeftY, CameraLeftZ, CameraViewX, CameraViewY, CameraViewZ, DegToRad(angDeg));
  VectorNormalize(CameraLeftX, CameraLeftY, CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);

  RotateVector(CameraUpX, CameraUpY, CameraUpZ, CameraViewX, CameraViewY, CameraViewZ, DegToRad(angDeg));
  VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);
end;

{----- TMyOpenGL.OrbitUp -----}
procedure TMyOpenGL.OrbitUp(const dist, angDeg : double);
var
  fixedX : double;
  fixedY : double;
  fixedZ : double;
  vecX : double;
  vecY : double;
  vecZ : double;
begin
  ProjectPoint(CameraPosX, CameraPosY, CameraPosZ, CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ, dist, fixedX, fixedY, fixedZ);
  vecX := CameraPosX - fixedX;
  vecY := CameraPosY - fixedY;
  vecZ := CameraPosZ - fixedZ;
  RotateVector(vecX, vecY, vecZ, CameraLeftX, CameraLeftY, CameraLeftZ, DegToRad(-angDeg));
  CameraPosX := vecX + fixedX;
  CameraPosY := vecY + fixedY;
  CameraPosZ := vecZ + fixedZ;
  RotateDown(angDeg);
end;

{----- TMyOpenGL.OrbitDown -----}
procedure TMyOpenGL.OrbitDown(const dist, angDeg : double);
var
  fixedX : double;
  fixedY : double;
  fixedZ : double;
  vecX : double;
  vecY : double;
  vecZ : double;
begin
  ProjectPoint(CameraPosX, CameraPosY, CameraPosZ, CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ, dist, fixedX, fixedY, fixedZ);
  vecX := CameraPosX - fixedX;
  vecY := CameraPosY - fixedY;
  vecZ := CameraPosZ - fixedZ;
  RotateVector(vecX, vecY, vecZ, CameraLeftX, CameraLeftY, CameraLeftZ, DegToRad(angDeg));
  CameraPosX := vecX + fixedX;
  CameraPosY := vecY + fixedY;
  CameraPosZ := vecZ + fixedZ;
  RotateUp(angDeg);
end;

{----- TMyOpenGL.OrbitLeft -----}
procedure TMyOpenGL.OrbitLeft(const dist, angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
var
  fixedX : double;
  fixedY : double;
  fixedZ : double;
  vecX : double;
  vecY : double;
  vecZ : double;
begin
  ProjectPoint(CameraPosX, CameraPosY, CameraPosZ, CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ, dist, fixedX, fixedY, fixedZ);
  vecX := CameraPosX - fixedX;
  vecY := CameraPosY - fixedY;
  vecZ := CameraPosZ - fixedZ;
  if not forceUpVector then RotateVector(vecX, vecY, vecZ, CameraUpX, CameraUpY, CameraUpZ, DegToRad(angDeg))
                       else RotateVector(vecX, vecY, vecZ, fUpX, fUpY, fUpZ, DegToRad(angDeg));
  CameraPosX := vecX + fixedX;
  CameraPosY := vecY + fixedY;
  CameraPosZ := vecZ + fixedZ;
  RotateRight(angDeg, forceUpVector, fUpX, fUpY, fUpZ);
end;

{----- TMyOpenGL.OrbitRight -----}
procedure TMyOpenGL.OrbitRight(const dist, angDeg : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
var
  fixedX : double;
  fixedY : double;
  fixedZ : double;
  vecX : double;
  vecY : double;
  vecZ : double;
begin
  ProjectPoint(CameraPosX, CameraPosY, CameraPosZ, CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ, dist, fixedX, fixedY, fixedZ);
  vecX := CameraPosX - fixedX;
  vecY := CameraPosY - fixedY;
  vecZ := CameraPosZ - fixedZ;
  if not forceUpVector then RotateVector(vecX, vecY, vecZ, CameraUpX, CameraUpY, CameraUpZ, DegToRad(-angDeg))
                       else RotateVector(vecX, vecY, vecZ, fUpX, fUpY, fUpZ, DegToRad(-angDeg));
  CameraPosX := vecX + fixedX;
  CameraPosY := vecY + fixedY;
  CameraPosZ := vecZ + fixedZ;
  RotateLeft(angDeg, forceUpVector, fUpX, fUpY, fUpZ);
end;

{----- TMyOpenGL.LookAt -----}
procedure TMyOpenGL.LookAt(const x, y, z : double; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
begin
  if SameValue(AngleBetweenTwoVectors(CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ, CameraPosX, CameraPosY, CameraPosZ, x, y, z), 0) then exit;

  VectorNormalize(x - CameraPosX, y - CameraPosY, z - CameraPosZ, CameraViewX, CameraViewY, CameraViewZ);
  if not forceUpVector then VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraUpX, CameraUpY, CameraUpZ, CameraLeftX, CameraLeftY, CameraLeftZ)
                       else VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, fUpX, fUpY, fUpZ, CameraLeftX, CameraLeftY, CameraLeftZ);
  VectorNormalize(-CameraLeftX, -CameraLeftY, -CameraLeftZ, CameraLeftX, CameraLeftY, CameraLeftZ);
  VectorCrossProduct(CameraViewX, CameraViewY, CameraViewZ, CameraLeftX, CameraLeftY, CameraLeftZ, CameraUpX, CameraUpY, CameraUpZ);
  VectorNormalize(CameraUpX, CameraUpY, CameraUpZ, CameraUpX, CameraUpY, CameraUpZ);
end;

{----- TMyOpenGL.LookAt -----}
procedure TMyOpenGL.LookAt(const v : TVector3f; const forceUpVector : boolean = false; const fUpX : double = 0; const fUpY : double = 0; const fUpZ : double = 0);
begin
  LookAt(v.X, v.Y, v.Z, forceUpVector, fUpX, fUpY, fUpZ);
end;

{----- TMyOpenGL.WorldToScreen -----}
function TMyOpenGL.WorldToScreen(const x, y, z : double; var nX, nY : integer) : boolean;
var
  vX : double;
  vY : double;
  vZ : double;
begin
  result := false;
  if not Rendering then Exit;

  if gluProject(x, y, z, ModelViewMatrix, ProjectionMatrix, ViewPortMatrix, @vX, @vY, @vZ) = 1 then result := true;
  nx := Floor(vX);
  ny := Floor(vY);
end;

{----- TMyOpenGL.WorldToScreen -----}
function TMyOpenGL.WorldToScreen(const v : TVector3f; var nX, nY : integer) : boolean;
begin
  result := WorldToScreen(v.X, v.Y, v.Z, nX, nY);
end;

{----- TMyOpenGL.RenderingBegin -----}
procedure TMyOpenGL.RenderingBegin;
var
  r : single;
  g : single;
  b : single;
  w : single;

  // antialising setting
  procedure SetAntiAliasing;
  begin
    with AntiAliasing do
      if Enable then // best but slowest
      begin
        glEnable(GL_SMOOTH);
        glEnable(GL_LINE_SMOOTH);
        glEnable(GL_POINT_SMOOTH);
        glEnable(GL_POLYGON_SMOOTH);
        glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
        glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
        glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
      end
      else // worst but fastest
      begin
        glDisable(GL_SMOOTH);
        glDisable(GL_LINE_SMOOTH);
        glDisable(GL_POINT_SMOOTH);
        glDisable(GL_POLYGON_SMOOTH);
        glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
        glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
        glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
      end;
  end;

  // blend setting
  procedure SetBlend;
  begin
    with Blend do
      if Enable then
      begin
        glEnable(GL_BLEND);
        glBlendFunc(FuncSrc, FuncDest);
      end
      else glDisable(GL_BLEND);
  end;

  // Depth buffer setting
  procedure SetDepthTest;
  begin
    with DepthTest do
      if Enable then
      begin
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(Func);
      end
      else glDisable(GL_DEPTH_TEST);
  end;

  // face culling setting
  procedure SetFaceCulling;
  begin
    with FaceCulling do
      if Enable then
      begin
        glEnable(GL_CULL_FACE);
        glCullFace(Func);
      end
      else glDisable(GL_CULL_FACE);
  end;

  // lights setting
  procedure SetLights;
  var
    i : integer;
    pos : array[0..3] of single;
    dir : array[0..2] of single;
  begin
    with Lights do
      if Enable then
      begin
        glEnable(GL_LIGHTING);
        glShadeModel(Shading);
        glEnable(GL_COLOR_MATERIAL);
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @GlobalAmbient);

        // settings all 8 lights
        for i := 0 to 7 do
          with Light[i] do
            if Enable then
            begin
              pos[0] := Position.X;
              pos[1] := Position.Y;
              pos[2] := Position.Z;
              pos[3] := 1;
              glLightfv(Func, GL_POSITION, @pos);

              dir[0] := Direction.X;
              dir[1] := Direction.Y;
              dir[2] := Direction.Z;
              glLightfv(Func, GL_SPOT_DIRECTION, @dir);

              glLightfv(Func, GL_AMBIENT,  @Ambient);
              glLightfv(Func, GL_DIFFUSE,  @Diffuse);
              glLightfv(Func, GL_SPECULAR, @Specular);

              glLightf(Func, GL_SPOT_CUTOFF,           SpotCutOff);
              glLightf(Func, GL_SPOT_EXPONENT,         SpotExponent);
              glLightf(Func, GL_CONSTANT_ATTENUATION,  Const_Atten);
              glLightf(Func, GL_LINEAR_ATTENUATION,    Linea_Atten);
              glLightf(Func, GL_QUADRATIC_ATTENUATION, Quadr_Atten);

              glEnable(Func);
            end
            else glDisable(Func);
      end
      else glDisable(GL_LIGHTING);
  end;

  // nastaven mlhy
  procedure SetFog;
  var
    GlColor : array[0..3] of single;
  begin
    with Fog do
      if Enable then
      begin
        glFogi(GL_FOG_MODE, Mode);
        glFogf(GL_FOG_DENSITY, Density);
        glFogf(GL_FOG_START, FogStart);
        glFogf(GL_FOG_END, FogEnd);
        glFogf(GL_FOG_INDEX, FogIndex);
        TColorToGLColor(Color, 1.0, GlColor[0], GlColor[1], GlColor[2], GlColor[3]);
        glFogfv(GL_FOG_COLOR, @GlColor);
        glFogi(GL_FOG_COORD_SRC, CoordSrc);
        glHint(GL_FOG_HINT, GL_DONT_CARE);
        glEnable(GL_FOG);
      end
      else glDisable(GL_FOG);
  end;

begin
  if Rendering then Exit;
  Rendering := True;

  // check canvas object size
  if (CanvasWidth <> Canvas.ClientWidth) or (CanvasHeight <> Canvas.ClientHeight) then ReCreate;

  // OpenGL activation for multiple classes on one form
  ActivateRenderingContext(DC, RC);

  // settings OpenGl function
  SetAntiAliasing;
  SetBlend;
  SetDepthTest;
  SetFaceCulling;
  SetLights;
  SetFog;

  // canvas reset
  TColorToGLColor(BackgroundColor, 1.0, r, g, b, w);
  glClearColor(r, g, b, w);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT); // delete depth buffer

  // camera and scene transformation
  ApplyTransformation;
end;

{----- TMyOpenGL.RenderingToOrthoCameraLayer -----}
procedure TMyOpenGL.RenderingToOrthoCameraLayer;
begin
  if not Rendering then Exit;

  // turn on ortho 2D layer of camera
  if Camera2DLayerActive then exit;
  Camera2DLayerActive := true;

  glDisable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);
  glDisable(GL_LIGHTING);
  glDisable(GL_FOG);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  glOrtho(0, CanvasWidth, 0, CanvasHeight, -1, 1); // setting ortho projection

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
end;

{----- TMyOpenGL.RenderingEnd -----}
procedure TMyOpenGL.RenderingEnd;
begin
  if not Rendering then Exit;

  // if ortho 2D layer active, we must restrore some OpenGl functions
  if Camera2DLayerActive then
  begin
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
  end;
  Camera2DLayerActive := false;

  SwapBuffers(DC);
  Rendering := False;
end;

{----- TMyOpenGL.SetGLColor -----}
procedure TMyOpenGL.SetGLColor(const color : TColor; const alpha : single);
var
  r : single;
  g : single;
  b : single;
  w : single;
begin
  if not Rendering then Exit;
  TColorToGLColor(color, alpha, r, g, b, w);
  glColor4f(r, g, b, w);
end;

{----- TMyOpenGL.SetGLPixelSize -----}
procedure TMyOpenGL.SetGLPixelSize(const size : single);
begin
  if not Rendering then Exit;
  glPointSize(size);
end;

{----- TMyOpenGL.SetGLLineWidth -----}
procedure TMyOpenGL.SetGLLineWidth(const width : single);
begin
  if not Rendering then Exit;
  glLineWidth(width);
end;

{----- TMyOpenGL.SetGLLineStyle -----}
procedure TMyOpenGL.SetGLLineStyle(const factor : Integer; const pattern : word);
begin
  if not Rendering then Exit;
  glEnable(GL_LINE_STIPPLE);
  glLineStipple(factor, pattern);
end;

{----- TMyOpenGL.SetGLLineStyle -----}
procedure TMyOpenGL.SetGLLineStyle(const style : TLineStyle; const enlarge : Byte = 2);
begin
  if not Rendering then Exit;
  case style of
    lsSolid      : SetGLLineStyleEnd;
    lsDash       : SetGLLineStyle(3 * enlarge, $5555);
    lsDashDot    : SetGLLineStyle(1 * enlarge, $6F6F);
    lsDashDotDot : SetGLLineStyle(2 * enlarge, $EAEA);
    lsDot        : SetGLLineStyle(1 * enlarge, $5555);
  end;
end;

{----- TMyOpenGL.SetGLLineStyleEnd -----}
procedure TMyOpenGL.SetGLLineStyleEnd;
begin
  if not Rendering then Exit;
  glDisable(GL_LINE_STIPPLE);
end;

{----- TMyOpenGL.BeginPixels -----}
procedure TMyOpenGL.BeginPixels;
begin
  if not Rendering then Exit;
  glBegin(GL_POINTS);
end;

{----- TMyOpenGL.EndPixels -----}
procedure TMyOpenGL.EndPixels;
begin
  if not Rendering then Exit;
  glEnd;
end;

{----- TMyOpenGL.BeginLines -----}
procedure TMyOpenGL.BeginLines;
begin
  if not Rendering then Exit;
  glBegin(GL_LINES);
end;

{----- TMyOpenGL.BeginLineStrip -----}
procedure TMyOpenGL.BeginLineStrip;
begin
  if not Rendering then Exit;
  glBegin(GL_LINE_STRIP);
end;

{----- TMyOpenGL.BeginLineLoop -----}
procedure TMyOpenGL.BeginLineLoop;
begin
  if not Rendering then Exit;
  glBegin(GL_LINE_LOOP);
end;

{----- TMyOpenGL.EndLines -----}
procedure TMyOpenGL.EndLines;
begin
  if not Rendering then Exit;
  glEnd;
end;

{----- TMyOpenGL.BeginTriangles -----}
procedure TMyOpenGL.BeginTriangles;
begin
  if not Rendering then Exit;
  glBegin(GL_TRIANGLES);
end;

{----- TMyOpenGL.BeginTriangleStrip -----}
procedure TMyOpenGL.BeginTriangleStrip;
begin
  if not Rendering then Exit;
  glBegin(GL_TRIANGLE_STRIP);
end;

{----- TMyOpenGL.BeginTriangleFan -----}
procedure TMyOpenGL.BeginTriangleFan;
begin
  if not Rendering then Exit;
  glBegin(GL_TRIANGLE_FAN);
end;

{----- TMyOpenGL.EndTriangles -----}
procedure TMyOpenGL.EndTriangles;
begin
  if not Rendering then Exit;
  glEnd;
end;

{----- TMyOpenGL.BeginQuads -----}
procedure TMyOpenGL.BeginQuads;
begin
  if not Rendering then Exit;
  glBegin(GL_QUADS);
end;

{----- TMyOpenGL.BeginQuadStrip -----}
procedure TMyOpenGL.BeginQuadStrip;
begin
  if not Rendering then Exit;
  glBegin(GL_QUAD_STRIP);
end;

{----- TMyOpenGL.EndQuads -----}
procedure TMyOpenGL.EndQuads;
begin
  if not Rendering then Exit;
  glEnd;
end;

{----- TMyOpenGL.BeginPolygon -----}
procedure TMyOpenGL.BeginPolygon;
begin
  if not Rendering then Exit;
  glBegin(GL_POLYGON);
end;

{----- TMyOpenGL.EndPolygon -----}
procedure TMyOpenGL.EndPolygon;
begin
  if not Rendering then Exit;
  glEnd;
end;

{----- TMyOpenGL.DrawVertex3f -----}
procedure TMyOpenGL.DrawVertex3f(const x, y, z : double);
begin
  if not Rendering then Exit;
  glVertex3f(x, y, z);
end;

{----- TMyOpenGL.DrawVertex3f -----}
procedure TMyOpenGL.DrawVertex3f(const v : TVector3f);
begin
  if not Rendering then Exit;
  DrawVertex3f(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.DrawVertexArray3f -----}
procedure TMyOpenGL.DrawVertexArray3f(const p : TVectorArray3f);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(p) - 1 do DrawVertex3f(p[i]);
end;

{----- TMyOpenGL.DrawNormal3f -----}
procedure TMyOpenGL.DrawNormal3f(const x, y, z : double);
begin
  if not Rendering then Exit;
  glNormal3f(x, y, z);
end;

{----- TMyOpenGL.DrawNormal3f -----}
procedure TMyOpenGL.DrawNormal3f(const v : TVector3f);
begin
  if not Rendering then Exit;
  DrawNormal3f(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.DrawNormalArray3f -----}
procedure TMyOpenGL.DrawNormalArray3f(const p : TVectorArray3f);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(p) - 1 do DrawNormal3f(p[i]);
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TPixelObject);
var
  i : integer;
  a : integer;
  v : TVector3f;
begin
  if not Rendering then Exit;

  if Lights.Enable then glDisable(GL_LIGHTING);

  with obj do
  begin
    if obj.Visible = false then exit;

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLPixelSize(PixelSize);
        SetGLColor(Color, Alpha);
        BeginPixels;

        for a := 0 to length(Vertices) - 1 do
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;

        EndPixels;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TLineObject);
var
  i : integer;
  a : integer;
  v : TVector3f;
begin
  if not Rendering then Exit;

  with obj do
  begin
    if obj.Visible = false then exit;

    if Lights.Enable then glDisable(GL_LIGHTING);

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLLineWidth(LineWidth);
        SetGLLineStyle(LineStyle);
        SetGLColor(Color, Alpha);

        if Closed then BeginLineLoop
                  else BeginLineStrip;

        for a := 0 to length(Vertices) - 1 do
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;

        EndLines;
        SetGLLineStyleEnd;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TTrianglesObject; const showNormals : boolean = false; const NormalLen : single = 1);
var
  i : integer;
  a : integer;
  v : TVector3f;
  n : TVector3f;
begin
  if not Rendering then Exit;

  with obj do
  begin
    if Visible = false then exit;

    if IgnoreLight then
      if Lights.Enable then glDisable(GL_LIGHTING);

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLColor(Color, Alpha);
        BeginTriangles;

        for a := 0 to length(Vertices) - 1 do
        begin
          // normal
          if a < length(Normals) then
            with Normals[a] do
            begin
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
              VectorNormalize(n, n);
              DrawNormal3f(n);
            end;

          // vertex
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;
        end;

        EndTriangles;

        // drawing normals
        if showNormals then
        begin
          SetGLLineStyle(lsSolid);
          SetGLLineWidth(1);
          BeginLines;

          for a := 0 to length(Vertices) - 1 do
          begin
            // normal
            if a < length(Normals) then
              with Normals[a] do
              begin
                VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
                VectorNormalize(n, n);
              end;

            // vertex
            with Vertices[a] do
            begin
              v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            end;

            DrawVertex3f(v);
            DrawVertex3f(v.X + n.X * NormalLen, v.Y + n.Y * NormalLen, v.Z + n.Z * NormalLen);
          end;

          EndLines;
        end;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TTriangleStripObject; const showNormals : boolean = false; const NormalLen : single = 1);
var
  i : integer;
  a : integer;
  v : TVector3f;
  n : TVector3f;
begin
  if not Rendering then Exit;

  with obj do
  begin
    if Visible = false then exit;

    if IgnoreLight then
      if Lights.Enable then glDisable(GL_LIGHTING);

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLColor(Color, Alpha);
        BeginTriangleStrip;

        for a := 0 to length(Vertices) - 1 do
        begin
          // normal
          if a < length(Normals) then
            with Normals[a] do
            begin
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
              VectorNormalize(n, n);
              DrawNormal3f(n);
            end;

          // vertex
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;
        end;

        EndTriangles;

        // drawing normals
        if showNormals then
        begin
          SetGLLineStyle(lsSolid);
          SetGLLineWidth(1);
          BeginLines;

          for a := 0 to length(Vertices) - 1 do
          begin
            // normal
            if a < length(Normals) then
              with Normals[a] do
              begin
                VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
                VectorNormalize(n, n);
              end;

            // vertex
            with Vertices[a] do
            begin
              v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            end;

            DrawVertex3f(v);
            DrawVertex3f(v.X + n.X * NormalLen, v.Y + n.Y * NormalLen, v.Z + n.Z * NormalLen);
          end;

          EndLines;
        end;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TTriangleFanObject; const showNormals : boolean = false; const NormalLen : single = 1);
var
  i : integer;
  a : integer;
  v : TVector3f;
  n : TVector3f;
begin
  if not Rendering then Exit;

  with obj do
  begin
    if Visible = false then exit;

    if IgnoreLight then
      if Lights.Enable then glDisable(GL_LIGHTING);

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLColor(Color, Alpha);
        BeginTriangleFan;

        for a := 0 to length(Vertices) - 1 do
        begin
          // normal
          if a < length(Normals) then
            with Normals[a] do
            begin
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
              VectorNormalize(n, n);
              DrawNormal3f(n);
            end;

          // vertex
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;
        end;

        EndTriangles;

        // drawing normals
        if showNormals then
        begin
          SetGLLineStyle(lsSolid);
          SetGLLineWidth(1);
          BeginLines;

          for a := 0 to length(Vertices) - 1 do
          begin
            // normal
            if a < length(Normals) then
              with Normals[a] do
              begin
                VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
                VectorNormalize(n, n);
              end;

            // vertex
            with Vertices[a] do
            begin
              v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            end;

            DrawVertex3f(v);
            DrawVertex3f(v.X + n.X * NormalLen, v.Y + n.Y * NormalLen, v.Z + n.Z * NormalLen);
          end;

          EndLines;
        end;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TQuadsObject; const showNormals : boolean = false; const NormalLen : single = 1);
var
  i : integer;
  a : integer;
  v : TVector3f;
  n : TVector3f;
begin
  if not Rendering then Exit;

  with obj do
  begin
    if Visible = false then exit;

    if IgnoreLight then
      if Lights.Enable then glDisable(GL_LIGHTING);

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLColor(Color, Alpha);
        BeginQuads;

        for a := 0 to length(Vertices) - 1 do
        begin
          // normal
          if a < length(Normals) then
            with Normals[a] do
            begin
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
              VectorNormalize(n, n);
              DrawNormal3f(n);
            end;

          // vertex
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;
        end;

        EndQuads;

        // drawing normals
        if showNormals then
        begin
          SetGLLineStyle(lsSolid);
          SetGLLineWidth(1);
          BeginLines;

          for a := 0 to length(Vertices) - 1 do
          begin
            // normal
            if a < length(Normals) then
              with Normals[a] do
              begin
                VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
                VectorNormalize(n, n);
              end;

            // vertex
            with Vertices[a] do
            begin
              v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            end;

            DrawVertex3f(v);
            DrawVertex3f(v.X + n.X * NormalLen, v.Y + n.Y * NormalLen, v.Z + n.Z * NormalLen);
          end;

          EndLines;
        end;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TQuadStripObject; const showNormals : boolean = false; const NormalLen : single = 1);
var
  i : integer;
  a : integer;
  v : TVector3f;
  n : TVector3f;
begin
  if not Rendering then Exit;

  with obj do
  begin
    if Visible = false then exit;

    if IgnoreLight then
      if Lights.Enable then glDisable(GL_LIGHTING);

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLColor(Color, Alpha);
        BeginQuadStrip;

        for a := 0 to length(Vertices) - 1 do
        begin
          // normal
          if a < length(Normals) then
            with Normals[a] do
            begin
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
              VectorNormalize(n, n);
              DrawNormal3f(n);
            end;

          // vertex
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;
        end;

        EndQuads;

        // drawing normals
        if showNormals then
        begin
          SetGLLineStyle(lsSolid);
          SetGLLineWidth(1);
          BeginLines;

          for a := 0 to length(Vertices) - 1 do
          begin
            // normal
            if a < length(Normals) then
              with Normals[a] do
              begin
                VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
                VectorNormalize(n, n);
              end;

            // vertex
            with Vertices[a] do
            begin
              v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            end;

            DrawVertex3f(v);
            DrawVertex3f(v.X + n.X * NormalLen, v.Y + n.Y * NormalLen, v.Z + n.Z * NormalLen);
          end;

          EndLines;
        end;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObject -----}
procedure TMyOpenGL.DrawObject(const obj : TPolygonObject; const showNormals : boolean = false; const NormalLen : single = 1);
var
  i : integer;
  a : integer;
  v : TVector3f;
  n : TVector3f;
begin
  if not Rendering then Exit;

  with obj do
  begin
    if Visible = false then exit;

    if IgnoreLight then
      if Lights.Enable then glDisable(GL_LIGHTING);

    DrawNormal3f(0, 0, 0);
    for i := 0 to length(Structure) - 1 do
      with Structure[i] do
      begin
        SetGLColor(Color, Alpha);
        BeginPolygon;

        for a := 0 to length(Vertices) - 1 do
        begin
          // normal
          if a < length(Normals) then
            with Normals[a] do
            begin
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
              VectorNormalize(n, n);
              DrawNormal3f(n);
            end;

          // vertex
          with Vertices[a] do
          begin
            v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
            VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            DrawVertex3f(v);
          end;
        end;

        EndPolygon;

        // drawing normals
        if showNormals then
        begin
          SetGLLineStyle(lsSolid);
          SetGLLineWidth(1);
          BeginLines;

          for a := 0 to length(Vertices) - 1 do
          begin
            // normal
            if a < length(Normals) then
              with Normals[a] do
              begin
                VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), X, Y, Z, n.X, n.Y, n.Z);
                VectorNormalize(n, n);
              end;

            // vertex
            with Vertices[a] do
            begin
              v := MakeVector3f(X * Scale.X + Position.X, Y * Scale.Y + Position.Y, Z * Scale.Z + Position.Z);
              VectorRotate(DegToRad(Rotation.X), DegToRad(Rotation.Y), DegToRad(Rotation.Z), v.X, v.Y, v.Z, Position.X, Position.Y, Position.Z, v.X, v.Y, v.Z);
            end;

            DrawVertex3f(v);
            DrawVertex3f(v.X + n.X * NormalLen, v.Y + n.Y * NormalLen, v.Z + n.Z * NormalLen);
          end;

          EndLines;
        end;
      end;

    if Lights.Enable then glEnable(GL_LIGHTING);
  end;
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TPixelObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TLineObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TTrianglesObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TTriangleStripObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TTriangleFanObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TQuadsObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TQuadStripObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.DrawObjectArray -----}
procedure TMyOpenGL.DrawObjectArray(const objArray : TPolygonObjectArray);
var
  i : integer;
begin
  if not Rendering then Exit;
  for i := 0 to length(objArray) - 1 do DrawObject(objArray[i]);
end;

{----- TMyOpenGL.GetLookAtDistVector -----}
procedure TMyOpenGL.GetLookAtDistVector(const Dist : double; var x, y, z : double);
begin
  ProjectPoint(CameraPosX, CameraPosY, CameraPosZ, CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ, Dist, x, y, z);
end;

{----- TMyOpenGL.GetLookAtDistVector -----}
function TMyOpenGL.GetLookAtDistVector(const Dist : double) : TVector3f;
var
  x : double;
  y : double;
  z : double;
begin
  GetLookAtDistVector(Dist, x, y, z);
  Result := MakeVector3f(x, y, z);
end;

{----- TMyOpenGL.VectorUpperLowerSide -----}
function TMyOpenGL.VectorUpperLowerSide(const x, y, z : double) : double;
var
  planeV1 : TVector3f;
  planeV2 : TVector3f;
  planeV3 : TVector3f;
  crossV  : TVector3f;
  vec     : TVector3f;
begin
  // rovina
  planeV1 := MakeVector3f(CameraPosX, CameraPosY, CameraPosZ);
  planeV2 := MakeVector3f(CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ);
  planeV3 := MakeVector3f(CameraPosX + CameraLeftX, CameraPosY + CameraLeftY, CameraPosZ + CameraLeftZ);

  VectorSubtract(x, y, z, planeV1.X, planeV1.Y, planeV1.Z, vec.X, vec.Y, vec.Z);
  VectorCrossProduct(planeV2.X - planeV1.X, planeV2.Y - planeV1.Y, planeV2.Z - planeV1.Z, planeV3.X - planeV1.X, planeV3.Y - planeV1.Y, planeV3.Z - planeV1.Z, crossV.X, crossV.Y, crossV.Z);

  Result := VectorDotProduct(vec, crossV); // positive = upper, negative = lower, zero = same plane
end;

{----- TMyOpenGL.VectorUpperLowerSide -----}
function TMyOpenGL.VectorUpperLowerSide(const v : TVector3f) : double;
begin
  Result := VectorUpperLowerSide(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.VectorLeftRightSide -----}
function TMyOpenGL.VectorLeftRightSide(const x, y, z : double) : double;
var
  planeV1 : TVector3f;
  planeV2 : TVector3f;
  planeV3 : TVector3f;
  crossV  : TVector3f;
  vec     : TVector3f;
begin
  // rovina
  planeV1 := MakeVector3f(CameraPosX, CameraPosY, CameraPosZ);
  planeV2 := MakeVector3f(CameraPosX + CameraViewX, CameraPosY + CameraViewY, CameraPosZ + CameraViewZ);
  planeV3 := MakeVector3f(CameraPosX + CameraUpX, CameraPosY + CameraUpY, CameraPosZ + CameraUpZ);

  VectorSubtract(x, y, z, planeV1.X, planeV1.Y, planeV1.Z, vec.X, vec.Y, vec.Z);
  VectorCrossProduct(planeV2.X - planeV1.X, planeV2.Y - planeV1.Y, planeV2.Z - planeV1.Z, planeV3.X - planeV1.X, planeV3.Y - planeV1.Y, planeV3.Z - planeV1.Z, crossV.X, crossV.Y, crossV.Z);

  Result := -VectorDotProduct(vec, crossV); // positive = left, negative = right, zero = same plane
end;

{----- TMyOpenGL.VectorLeftRightSide -----}
function TMyOpenGL.VectorLeftRightSide(const v : TVector3f) : double;
begin
  Result := VectorLeftRightSide(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.VectorFrontBehindSide -----}
function TMyOpenGL.VectorFrontBehindSide(const x, y, z : double) : double;
var
  planeV1 : TVector3f;
  planeV2 : TVector3f;
  planeV3 : TVector3f;
  crossV  : TVector3f;
  vec     : TVector3f;
begin
  // rovina
  planeV1 := MakeVector3f(CameraPosX, CameraPosY, CameraPosZ);
  planeV2 := MakeVector3f(CameraPosX + CameraLeftX, CameraPosY + CameraLeftY, CameraPosZ + CameraLeftZ);
  planeV3 := MakeVector3f(CameraPosX + CameraUpX, CameraPosY + CameraUpY, CameraPosZ + CameraUpZ);

  VectorSubtract(x, y, z, planeV1.X, planeV1.Y, planeV1.Z, vec.X, vec.Y, vec.Z);
  VectorCrossProduct(planeV2.X - planeV1.X, planeV2.Y - planeV1.Y, planeV2.Z - planeV1.Z, planeV3.X - planeV1.X, planeV3.Y - planeV1.Y, planeV3.Z - planeV1.Z, crossV.X, crossV.Y, crossV.Z);

  Result := VectorDotProduct(vec, crossV); // positive = front, negative = behind, zero = same plane
end;

{----- TMyOpenGL.VectorFrontBehindSide -----}
function TMyOpenGL.VectorFrontBehindSide(const v : TVector3f) : double;
begin
  Result := VectorFrontBehindSide(v.X, v.Y, v.Z);
end;

{----- TMyOpenGL.GetMouseRay -----}
procedure TMyOpenGL.GetMouseRay(const mouseX, mouseY : integer; var vStart, vEnd : TVector3f);
var
  x : double;
  y : double;
  z : double;
begin
  gluUnProject(MouseX, CanvasHeight - mouseY, 1, ModelViewMatrix, ProjectionMatrix, ViewPortMatrix, @x, @y, @z);

  vStart := GetCameraPosition;
  vEnd   := MakeVector3f(x, y, z);
end;

{----- CreateLineObject_Axes -----}
procedure CreateLineObject_Axes(const length : double; var obj : TLineObject);
begin
  with obj do
  begin
    SetLength(Structure, 6);
    with Structure[0] do // axis +X
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, 0, 0);
      Vertices[1] := MakeVector3f(+length, 0, 0);
      Color       := clRed;
      Alpha       := 1.0;
      LineWidth   := 1;
      LineStyle   := lsDot;
      Closed      := false;
    end;
    with Structure[1] do // axis -X
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, 0, 0);
      Vertices[1] := MakeVector3f(-length, 0, 0);
      Color       := clMaroon;
      Alpha       := 1.0;
      LineWidth   := 1;
      LineStyle   := lsDot;
      Closed      := false;
    end;
    with Structure[2] do // axis +Y
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, 0, 0);
      Vertices[1] := MakeVector3f(0, +length, 0);
      Color       := clLime;
      Alpha       := 1.0;
      LineWidth   := 1;
      LineStyle   := lsDot;
      Closed      := false;
    end;
    with Structure[3] do // axis -Y
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, 0, 0);
      Vertices[1] := MakeVector3f(0, -length, 0);
      Color       := clGreen;
      Alpha       := 1.0;
      LineWidth   := 1;
      LineStyle   := lsDot;
      Closed      := false;
    end;
    with Structure[4] do // axis +Z
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, 0, 0);
      Vertices[1] := MakeVector3f(0, 0, +length);
      Color       := clBlue;
      Alpha       := 1.0;
      LineWidth   := 1;
      LineStyle   := lsDot;
      Closed      := false;
    end;
    with Structure[5] do // axis -Z
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, 0, 0);
      Vertices[1] := MakeVector3f(0, 0, -length);
      Color       := clNavy;
      Alpha       := 1.0;
      LineWidth   := 1;
      LineStyle   := lsDot;
      Closed      := false;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Grid -----}
procedure CreateLineObject_Grid(const step : double; const blocks : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
var
  i : integer;
begin
  with obj do
  begin
    SetLength(Structure, blocks * 2 + 2);
    for i := 0 to blocks do
    begin
      with Structure[2 * i] do // vertical lines
      begin
        SetLength(Vertices, 2);
        Vertices[0] := MakeVector3f((i * step) - (blocks * step / 2),               0 - (blocks * step / 2), 0);
        Vertices[1] := MakeVector3f((i * step) - (blocks * step / 2), (blocks * step) - (blocks * step / 2), 0);
        Color       := lColor;
        Alpha       := lAlpha;
        LineWidth   := lWidth;
        LineStyle   := lStyle;
        Closed      := false;
      end;
      with Structure[2 * i + 1] do // horizontal lines
      begin
        SetLength(Vertices, 2);
        Vertices[0] := MakeVector3f(               0 - (blocks * step / 2), (i * step) - (blocks * step / 2), 0);
        Vertices[1] := MakeVector3f((blocks * step)  - (blocks * step / 2), (i * step) - (blocks * step / 2), 0);
        Color       := lColor;
        Alpha       := lAlpha;
        LineWidth   := lWidth;
        LineStyle   := lStyle;
        Closed      := false;
      end;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Line -----}
procedure CreateLineObject_Line(const x1, y1, z1, x2, y2, z2 : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
begin
  with obj do
  begin
    SetLength(Structure, 1);
    with Structure[0] do
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(x1, y1, z1);
      Vertices[1] := MakeVector3f(x2, y2, z2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Arc -----}
procedure CreateLineObject_Arc(const radius, angleStart, angleSweep : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    SetLength(Structure, 1);
    angle := angleSweep / segments;
    with Structure[0] do
    begin
      SetLength(Vertices, segments + 1);
      for i := 0 to segments do
        Vertices[i] := MakeVector3f(radius * cos(DegToRad(angleStart + angle * i)), radius * sin(DegToRad(angleStart + angle * i)), 0);

      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := false;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Circle -----}
procedure CreateLineObject_Circle(const radius : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    SetLength(Structure, 1);
    angle := 360 / segments;
    with Structure[0] do
    begin
      SetLength(Vertices, segments + 1);
      for i := 0 to segments do
        Vertices[i] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), 0);

      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := false;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Cross -----}
procedure CreateLineObject_Cross(const size : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
begin
  with obj do
  begin
    SetLength(Structure, 3);
    with Structure[0] do // first part
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(-size/2, 0, 0);
      Vertices[1] := MakeVector3f(+size/2, 0, 0);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    with Structure[1] do // second part
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, -size/2, 0);
      Vertices[1] := MakeVector3f(0, +size/2, 0);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    with Structure[2] do // third part
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(0, 0, -size/2);
      Vertices[1] := MakeVector3f(0, 0, +size/2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Cube -----}
procedure CreateLineObject_Cube(const sizeX, sizeY, sizeZ : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
begin
  with obj do
  begin
    SetLength(Structure, 6);
    with Structure[0] do // front side
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(+sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(-sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[2] := MakeVector3f(-sizeX/2, +sizeY/2, -sizeZ/2);
      Vertices[3] := MakeVector3f(+sizeX/2, +sizeY/2, -sizeZ/2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := true;
    end;
    with Structure[1] do // back side
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(+sizeX/2, -sizeY/2, +sizeZ/2);
      Vertices[1] := MakeVector3f(-sizeX/2, -sizeY/2, +sizeZ/2);
      Vertices[2] := MakeVector3f(-sizeX/2, +sizeY/2, +sizeZ/2);
      Vertices[3] := MakeVector3f(+sizeX/2, +sizeY/2, +sizeZ/2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := true;
    end;
    with Structure[2] do
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(+sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(+sizeX/2, -sizeY/2, +sizeZ/2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    with Structure[3] do
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(-sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(-sizeX/2, -sizeY/2, +sizeZ/2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    with Structure[4] do
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(-sizeX/2, +sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(-sizeX/2, +sizeY/2, +sizeZ/2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    with Structure[5] do
    begin
      SetLength(Vertices, 2);
      Vertices[0] := MakeVector3f(+sizeX/2, +sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(+sizeX/2, +sizeY/2, +sizeZ/2);
      Color       := lColor;
      Alpha       := lAlpha;
      LineWidth   := lWidth;
      LineStyle   := lStyle;
      Closed      := false;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Cylinder -----}
procedure CreateLineObject_Cylinder(const radius, height : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    SetLength(Structure, 3 + segments);
    angle := 360 / segments;
    with Structure[0] do // front circle
    begin
      SetLength(Vertices, (segments + 1) * 2);
      for i := 0 to segments do
      begin
        Vertices[i * 2]     := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), -height/2);
        Vertices[i * 2 + 1] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), -height/2);
      end;
      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := false;
    end;
    with Structure[1] do // back circle
    begin
      SetLength(Vertices, (segments + 1) * 2);
      for i := 0 to segments do
      begin
        Vertices[i * 2]     := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), +height/2);
        Vertices[i * 2 + 1] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), +height/2);
      end;
      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := false;
    end;
    for i := 0 to segments do
      with Structure[2 + i] do // connection between front and back circle
      begin
        SetLength(Vertices, 2);
        Vertices[0] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), -height/2);
        Vertices[1] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), +height/2);
        Color       := lColor;
        Alpha       := lAlpha;
        LineWidth   := lWidth;
        LineStyle   := lStyle;
        Closed      := false;
      end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Pie -----}
procedure CreateLineObject_Pie(const radius : double; const segments : integer; const CutoutPercents : double; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    SetLength(Structure, 2 + segments);
    angle := (360 / 100 * CutoutPercents) / segments;
    with Structure[0] do // circle
    begin
      SetLength(Vertices, (segments + 1) * 2);
      for i := 0 to segments do
      begin
        Vertices[i * 2]     := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), 0);
        Vertices[i * 2 + 1] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), 0);
      end;
      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := false;
    end;
    for i := 0 to segments do
      with Structure[1 + i] do // connection between circle and center
      begin
        SetLength(Vertices, 2);
        Vertices[0] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), 0);
        Vertices[1] := MakeVector3f(0, 0, 0);
        Color       := lColor;
        Alpha       := lAlpha;
        LineWidth   := lWidth;
        LineStyle   := lStyle;
        Closed      := false;
      end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Ring -----}
procedure CreateLineObject_Ring(const radiusInner, radiusOuter : double; const segments : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    SetLength(Structure, 3 + segments);
    angle := 360 / segments;
    with Structure[0] do // inner circle
    begin
      SetLength(Vertices, (segments + 1) * 2);
      for i := 0 to segments do
      begin
        Vertices[i * 2]     := MakeVector3f(radiusInner * cos(DegToRad(angle * i)), radiusInner * sin(DegToRad(angle * i)), 0);
        Vertices[i * 2 + 1] := MakeVector3f(radiusInner * cos(DegToRad(angle * i)), radiusInner * sin(DegToRad(angle * i)), 0);
      end;
      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := false;
    end;
    with Structure[1] do // outer circle
    begin
      SetLength(Vertices, (segments + 1) * 2);
      for i := 0 to segments do
      begin
        Vertices[i * 2]     := MakeVector3f(radiusOuter * cos(DegToRad(angle * i)), radiusOuter * sin(DegToRad(angle * i)), 0);
        Vertices[i * 2 + 1] := MakeVector3f(radiusOuter * cos(DegToRad(angle * i)), radiusOuter * sin(DegToRad(angle * i)), 0);
      end;
      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := false;
    end;
    for i := 0 to segments do
      with Structure[2 + i] do // connection beween inner and outer circle
      begin
        SetLength(Vertices, 2);
        Vertices[0] := MakeVector3f(radiusInner * cos(DegToRad(angle * i)), radiusInner * sin(DegToRad(angle * i)), 0);
        Vertices[1] := MakeVector3f(radiusOuter * cos(DegToRad(angle * i)), radiusOuter * sin(DegToRad(angle * i)), 0);
        Color       := lColor;
        Alpha       := lAlpha;
        LineWidth   := lWidth;
        LineStyle   := lStyle;
        Closed      := false;
      end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- CreateLineObject_Star -----}
procedure CreateLineObject_Star(const radiusInner, radiusOuter : double; const tips : integer; const lWidth : single; const lStyle : TLineStyle; const lColor : TColor; const lAlpha : single; var obj : TLineObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    SetLength(Structure, 1);
    angle := 360 / tips;
    with Structure[0] do
    begin
      SetLength(Vertices, (tips + 1) * 2);
      for i := 0 to tips do
      begin
        Vertices[i * 2]     := MakeVector3f(radiusInner * cos(DegToRad(angle * i)), radiusInner * sin(DegToRad(angle * i)), 0);
        Vertices[i * 2 + 1] := MakeVector3f(radiusOuter * cos(DegToRad(angle * i + angle / 2)), radiusOuter * sin(DegToRad(angle * i + angle / 2)), 0);
      end;
      Color     := lColor;
      Alpha     := lAlpha;
      LineWidth := lWidth;
      LineStyle := lStyle;
      Closed    := true;
    end;
    Position := MakeVector3f(0, 0, 0);
    Rotation := MakeVector3f(0, 0, 0);
    Scale    := MakeVector3f(1, 1, 1);
    Visible  := true;
  end;
end;

{----- Create3DObject_Plane -----}
procedure Create3DObject_Plane(const height, width : double; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TQuadsObject);
begin
  with obj do
  begin
    if twoSides then SetLength(Structure, 2)
                else SetLength(Structure, 1);
    with Structure[0] do // front quad
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(+width/2, -height/2, 0);
      Vertices[1] := MakeVector3f(-width/2, -height/2, 0);
      Vertices[2] := MakeVector3f(-width/2, +height/2, 0);
      Vertices[3] := MakeVector3f(+width/2, +height/2, 0);
      Color := pFrontColor;
      Alpha := pAlpha;

      SetLength(Normals, 4);
      Normals[0] := MakeVector3f(0, 0, -1);
      Normals[1] := MakeVector3f(0, 0, -1);
      Normals[2] := MakeVector3f(0, 0, -1);
      Normals[3] := MakeVector3f(0, 0, -1);
    end;
    if twoSides then
      with Structure[1] do // back quad
      begin
        SetLength(Vertices, 4);
        Vertices[0] := MakeVector3f(-width/2, -height/2, 0);
        Vertices[1] := MakeVector3f(+width/2, -height/2, 0);
        Vertices[2] := MakeVector3f(+width/2, +height/2, 0);
        Vertices[3] := MakeVector3f(-width/2, +height/2, 0);
        Color := pBackColor;
        Alpha := pAlpha;

        SetLength(Normals, 4);
        Normals[0] := MakeVector3f(0, 0, 1);
        Normals[1] := MakeVector3f(0, 0, 1);
        Normals[2] := MakeVector3f(0, 0, 1);
        Normals[3] := MakeVector3f(0, 0, 1);
      end;
    Position    := MakeVector3f(0, 0, 0);
    Rotation    := MakeVector3f(0, 0, 0);
    Scale       := MakeVector3f(1, 1, 1);
    Visible     := true;
    IgnoreLight := false;
  end;
end;

{----- Create3DObject_Cube -----}
procedure Create3DObject_Cube(const sizeX, sizeY, sizeZ : double; const pFrontColor, pBackColor, pTopColor, pBottomColor, pLeftColor, pRightColor : TColor; const pAlpha : single; var obj : TQuadsObject);
begin
  with obj do
  begin
    SetLength(Structure, 6);
    with Structure[0] do // front quad
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(+sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(-sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[2] := MakeVector3f(-sizeX/2, +sizeY/2, -sizeZ/2);
      Vertices[3] := MakeVector3f(+sizeX/2, +sizeY/2, -sizeZ/2);
      Color := pFrontColor;
      Alpha := pAlpha;

      SetLength(Normals, 4);
      Normals[0] := MakeVector3f(0, 0, -1);
      Normals[1] := MakeVector3f(0, 0, -1);
      Normals[2] := MakeVector3f(0, 0, -1);
      Normals[3] := MakeVector3f(0, 0, -1);
    end;
    with Structure[1] do // back quad
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(-sizeX/2, -sizeY/2, +sizeZ/2);
      Vertices[1] := MakeVector3f(+sizeX/2, -sizeY/2, +sizeZ/2);
      Vertices[2] := MakeVector3f(+sizeX/2, +sizeY/2, +sizeZ/2);
      Vertices[3] := MakeVector3f(-sizeX/2, +sizeY/2, +sizeZ/2);
      Color := pBackColor;
      Alpha := pAlpha;

      SetLength(Normals, 4);
      Normals[0] := MakeVector3f(0, 0, 1);
      Normals[1] := MakeVector3f(0, 0, 1);
      Normals[2] := MakeVector3f(0, 0, 1);
      Normals[3] := MakeVector3f(0, 0, 1);
    end;
    with Structure[2] do // top quad
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(+sizeX/2, +sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(-sizeX/2, +sizeY/2, -sizeZ/2);
      Vertices[2] := MakeVector3f(-sizeX/2, +sizeY/2, +sizeZ/2);
      Vertices[3] := MakeVector3f(+sizeX/2, +sizeY/2, +sizeZ/2);
      Color := pTopColor;
      Alpha := pAlpha;

      SetLength(Normals, 4);
      Normals[0] := MakeVector3f(0, 1, 0);
      Normals[1] := MakeVector3f(0, 1, 0);
      Normals[2] := MakeVector3f(0, 1, 0);
      Normals[3] := MakeVector3f(0, 1, 0);
    end;
    with Structure[3] do // bottom quad
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(-sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(+sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[2] := MakeVector3f(+sizeX/2, -sizeY/2, +sizeZ/2);
      Vertices[3] := MakeVector3f(-sizeX/2, -sizeY/2, +sizeZ/2);
      Color := pBottomColor;
      Alpha := pAlpha;

      SetLength(Normals, 4);
      Normals[0] := MakeVector3f(0, -1, 0);
      Normals[1] := MakeVector3f(0, -1, 0);
      Normals[2] := MakeVector3f(0, -1, 0);
      Normals[3] := MakeVector3f(0, -1, 0);
    end;
    with Structure[4] do // left quad
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(+sizeX/2, -sizeY/2, +sizeZ/2);
      Vertices[1] := MakeVector3f(+sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[2] := MakeVector3f(+sizeX/2, +sizeY/2, -sizeZ/2);
      Vertices[3] := MakeVector3f(+sizeX/2, +sizeY/2, +sizeZ/2);
      Color := pLeftColor;
      Alpha := pAlpha;

      SetLength(Normals, 4);
      Normals[0] := MakeVector3f(1, 0, 0);
      Normals[1] := MakeVector3f(1, 0, 0);
      Normals[2] := MakeVector3f(1, 0, 0);
      Normals[3] := MakeVector3f(1, 0, 0);
    end;
    with Structure[5] do // right quad
    begin
      SetLength(Vertices, 4);
      Vertices[0] := MakeVector3f(-sizeX/2, -sizeY/2, -sizeZ/2);
      Vertices[1] := MakeVector3f(-sizeX/2, -sizeY/2, +sizeZ/2);
      Vertices[2] := MakeVector3f(-sizeX/2, +sizeY/2, +sizeZ/2);
      Vertices[3] := MakeVector3f(-sizeX/2, +sizeY/2, -sizeZ/2);
      Color := pRightColor;
      Alpha := pAlpha;

      SetLength(Normals, 4);
      Normals[0] := MakeVector3f(-1, 0, 0);
      Normals[1] := MakeVector3f(-1, 0, 0);
      Normals[2] := MakeVector3f(-1, 0, 0);
      Normals[3] := MakeVector3f(-1, 0, 0);
    end;
    Position    := MakeVector3f(0, 0, 0);
    Rotation    := MakeVector3f(0, 0, 0);
    Scale       := MakeVector3f(1, 1, 1);
    Visible     := true;
    IgnoreLight := false;
  end;
end;

{----- Create3DObject_Cylinder -----}
procedure Create3DObject_Cylinder(const radius, height : double; const segments : integer; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TQuadStripObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    if twoSides then SetLength(Structure, 2)
                else SetLength(Structure, 1);
    angle := 360 / segments;
    with Structure[0] do // outer side
    begin
      SetLength(Vertices, (segments + 1) * 2);
      SetLength(Normals, (segments + 1) * 2);
      for i := 0 to segments do
      begin
        Vertices[i * 2]     := MakeVector3f(radius * cos(DegToRad(-angle * i)), radius * sin(DegToRad(-angle * i)), -height/2);
        Vertices[i * 2 + 1] := MakeVector3f(radius * cos(DegToRad(-angle * i)), radius * sin(DegToRad(-angle * i)), +height/2);
        Normals[i * 2]      := MakeVector3f(radius * cos(DegToRad(-angle * i)), radius * sin(DegToRad(-angle * i)), 0);
        Normals[i * 2 + 1]  := MakeVector3f(radius * cos(DegToRad(-angle * i)), radius * sin(DegToRad(-angle * i)), 0);
      end;
      Color := pFrontColor;
      Alpha := pAlpha;
    end;
    if twoSides then
      with Structure[1] do // inner side
      begin
        SetLength(Vertices, (segments + 1) * 2);
        SetLength(Normals, (segments + 1) * 2);
        for i := 0 to segments do
        begin
          Vertices[i * 2]     := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), -height/2);
          Vertices[i * 2 + 1] := MakeVector3f(radius * cos(DegToRad(angle * i)), radius * sin(DegToRad(angle * i)), +height/2);
          Normals[i * 2]      := MakeVector3f(-radius * cos(DegToRad(angle * i)), -radius * sin(DegToRad(angle * i)), 0);
          Normals[i * 2 + 1]  := MakeVector3f(-radius * cos(DegToRad(angle * i)), -radius * sin(DegToRad(angle * i)), 0);
        end;
        Color := pBackColor;
        Alpha := pAlpha;
      end;
    Position    := MakeVector3f(0, 0, 0);
    Rotation    := MakeVector3f(0, 0, 0);
    Scale       := MakeVector3f(1, 1, 1);
    Visible     := true;
    IgnoreLight := false;
  end;
end;

{----- Create3DObject_Pie -----}
procedure Create3DObject_Pie(const radius : double; const segments : integer; const CutoutPercents : double; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TTriangleFanObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    if twoSides then SetLength(Structure, 2)
                else SetLength(Structure, 1);
    angle := (360 / 100 * CutoutPercents) / segments;
    with Structure[0] do // front side
    begin
      SetLength(Vertices, segments + 2);
      SetLength(Normals, segments + 2);
      Vertices[0] := MakeVector3f(0, 0, 0);
      Normals[0]  := MakeVector3f(0, 0, -1);
      for i := segments + 1 downto 1 do
      begin
        Vertices[segments + 2 - i] := MakeVector3f(radius * cos(DegToRad(angle * (i-1))), radius * sin(DegToRad(angle * (i-1))), 0);
        Normals[segments + 2 - i]  := MakeVector3f(0, 0, -1);
      end;
      Color := pFrontColor;
      Alpha := pAlpha;
    end;
    if twoSides then
      with Structure[1] do // back side
      begin
        SetLength(Vertices, segments + 2);
        SetLength(Normals, segments + 2);
        Vertices[0] := MakeVector3f(0, 0, 0);
        Normals[0]  := MakeVector3f(0, 0, 1);
        for i := 1 to segments + 1 do
        begin
          Vertices[i] := MakeVector3f(radius * cos(DegToRad(angle * (i-1))), radius * sin(DegToRad(angle * (i-1))), 0);
          Normals[i]  := MakeVector3f(0, 0, 1);
        end;
        Color := pBackColor;
        Alpha := pAlpha;
      end;
    Position    := MakeVector3f(0, 0, 0);
    Rotation    := MakeVector3f(0, 0, 0);
    Scale       := MakeVector3f(1, 1, 1);
    Visible     := true;
    IgnoreLight := false;
  end;
end;

{----- Create3DObject_Ring -----}
procedure Create3DObject_Ring(const radiusInner, radiusOuter : double; const segments : integer; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TQuadStripObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    if twoSides then SetLength(Structure, 2)
                else SetLength(Structure, 1);
    angle := 360 / segments;
    with Structure[0] do // front side
    begin
      SetLength(Vertices, (segments + 1) * 2);
      SetLength(Normals, (segments + 1) * 2);
      for i := 0 to segments do
      begin
        Vertices[i * 2]     := MakeVector3f(radiusInner * cos(DegToRad(-angle * i)), radiusInner * sin(DegToRad(-angle * i)), 0);
        Vertices[i * 2 + 1] := MakeVector3f(radiusOuter * cos(DegToRad(-angle * i)), radiusOuter * sin(DegToRad(-angle * i)), 0);
        Normals[i * 2]      := MakeVector3f(0, 0, -1);
        Normals[i * 2 + 1]  := MakeVector3f(0, 0, -1);
      end;
      Color := pFrontColor;
      Alpha := pAlpha;
    end;
    if twoSides then
      with Structure[1] do // back side
      begin
        SetLength(Vertices, (segments + 1) * 2);
        SetLength(Normals, (segments + 1) * 2);
        for i := 0 to segments do
        begin
          Vertices[i * 2]     := MakeVector3f(radiusInner * cos(DegToRad(angle * i)), radiusInner * sin(DegToRad(angle * i)), 0);
          Vertices[i * 2 + 1] := MakeVector3f(radiusOuter * cos(DegToRad(angle * i)), radiusOuter * sin(DegToRad(angle * i)), 0);
          Normals[i * 2]      := MakeVector3f(0, 0, 1);
          Normals[i * 2 + 1]  := MakeVector3f(0, 0, 1);
        end;
        Color := pBackColor;
        Alpha := pAlpha;
      end;
    Position    := MakeVector3f(0, 0, 0);
    Rotation    := MakeVector3f(0, 0, 0);
    Scale       := MakeVector3f(1, 1, 1);
    Visible     := true;
    IgnoreLight := false;
  end;
end;

{----- Create3DObject_Star -----}
procedure Create3DObject_Star(const radiusInner, radiusOuter : double; const tips : integer; const twoSides : boolean; const pFrontColor, pBackColor : TColor; const pAlpha : single; var obj : TPolygonObject);
var
  i : integer;
  angle : double;
begin
  with obj do
  begin
    if twoSides then SetLength(Structure, 2 + (tips * 2))
                else SetLength(Structure, 1 + tips);
    angle := 360 / tips;
    with Structure[0] do // front center side
    begin
      SetLength(Vertices, tips);
      SetLength(Normals, tips);
      for i := 0 to tips - 1 do
      begin
        Vertices[i] := MakeVector3f(radiusInner * cos(DegToRad(-angle * i)), radiusInner * sin(DegToRad(-angle * i)), 0);
        Normals[i]  := MakeVector3f(0, 0, -1);
      end;
      Color := pFrontColor;
      Alpha := pAlpha;
    end;
    for i := 0 to tips - 1 do
      with Structure[1 + i] do // front tip side
      begin
        SetLength(Vertices, 3);
        SetLength(Normals, 3);
        Vertices[0] := MakeVector3f(radiusInner * cos(DegToRad(-angle * i)), radiusInner * sin(DegToRad(-angle * i)), 0);
        Normals[0]  := MakeVector3f(0, 0, -1);
        Vertices[1] := MakeVector3f(radiusOuter * cos(DegToRad(-angle * i - angle / 2)), radiusOuter * sin(DegToRad(-angle * i - angle / 2)), 0);
        Normals[1]  := MakeVector3f(0, 0, -1);
        Vertices[2] := MakeVector3f(radiusInner * cos(DegToRad(-angle * (i+1))), radiusInner * sin(DegToRad(-angle * (i+1))), 0);
        Normals[2]  := MakeVector3f(0, 0, -1);
        Color := pFrontColor;
        Alpha := pAlpha;
      end;
    if twoSides then
    begin
      with Structure[1 + tips] do // back center side
      begin
        SetLength(Vertices, tips);
        SetLength(Normals, tips);
        for i := 0 to tips - 1 do
        begin
          Vertices[i] := MakeVector3f(radiusInner * cos(DegToRad(angle * i)), radiusInner * sin(DegToRad(angle * i)), 0);
          Normals[i]  := MakeVector3f(0, 0, 1);
        end;
        Color := pBackColor;
        Alpha := pAlpha;
      end;
      for i := 0 to tips - 1 do
        with Structure[2 + tips + i] do // back tip side
        begin
          SetLength(Vertices, 3);
          SetLength(Normals, 3);
          Vertices[0] := MakeVector3f(radiusInner * cos(DegToRad(angle * i)), radiusInner * sin(DegToRad(angle * i)), 0);
          Normals[0]  := MakeVector3f(0, 0, 1);
          Vertices[1] := MakeVector3f(radiusOuter * cos(DegToRad(angle * i + angle / 2)), radiusOuter * sin(DegToRad(angle * i + angle / 2)), 0);
          Normals[1]  := MakeVector3f(0, 0, 1);
          Vertices[2] := MakeVector3f(radiusInner * cos(DegToRad(angle * (i+1))), radiusInner * sin(DegToRad(angle * (i+1))), 0);
          Normals[2]  := MakeVector3f(0, 0, 1);
          Color := pBackColor;
          Alpha := pAlpha;
        end;
    end;
    Position    := MakeVector3f(0, 0, 0);
    Rotation    := MakeVector3f(0, 0, 0);
    Scale       := MakeVector3f(1, 1, 1);
    Visible     := true;
    IgnoreLight := false;
  end;
end;

{----- MakeVector3f -----}
procedure MakeVector3f(const x, y, z : double; var v : TVector3f);
begin
  v.X := x;
  v.Y := y;
  v.Z := z;
end;

{----- MakeVector3f -----}
function MakeVector3f(const x, y, z : double) : TVector3f;
begin
  MakeVector3f(x, y, z, result);
end;

{----- GLColorToTColor -----}
function GLColorToTColor(const r, g, b : single) : TColor;
begin
  Result := RGB(Trunc(255 * r), Trunc(255 * g), Trunc(255 * b));
end;

{----- TColorToGLColor -----}
procedure TColorToGLColor(const color : TColor; const alpha : single; var r, g, b, w : single);
begin
  r := GetRValue(color) / 255;
  g := GetGValue(color) / 255;
  b := GetBValue(color) / 255;
  w := alpha;
end;

{----- VectorEquals -----}
function VectorEquals(const x1, y1, z1, x2, y2, z2 : double; const epsilon : double = 0.00001) : boolean;
begin
  Result := SameValue(x1, x2, epsilon) and SameValue(y1, y2, epsilon) and SameValue(z1, z2, epsilon);
end;

{----- VectorEquals -----}
function VectorEquals(const v1, v2 : TVector3f; const epsilon : double = 0.00001) : boolean;
begin
  Result := VectorEquals(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, epsilon);
end;

{----- VectorDistance -----}
function VectorDistance(const x1, y1, z1, x2, y2, z2 : double) : double;
var
  dx : double;
  dy : double;
  dz : double;
begin
  dx := x2 - x1;
  dy := y2 - y1;
  dz := z2 - z1;
  Result := Sqrt(dx * dx + dy * dy + dz * dz);
end;

{----- VectorDistance -----}
function VectorDistance(const v1, v2 : TVector3f) : double;
begin
  Result := VectorDistance(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z);
end;

{----- VectorNormalize -----}
procedure VectorNormalize(const x, y, z : double; var vX, vY, vZ : double);
var
  d : extended;
begin
  d := Sqrt(Sqr(x)+Sqr(y)+Sqr(z));
  if d <> 0 then
  begin
    vX := x / d;
    vY := y / d;
    vZ := z / d;
  end
  else // non-valid result
  begin
    vX := 0;
    vY := 0;
    vZ := 0;
  end;
end;

{----- VectorNormalize -----}
procedure VectorNormalize(const v1 : TVector3f; var v2 : TVector3f);
begin
  VectorNormalize(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z);
end;

{----- VectorNormalize -----}
function VectorNormalize(const v : TVector3f) : TVector3f;
begin
  VectorNormalize(v.X, v.Y, v.Z, result.X, result.Y, result.Z);
end;

{----- VectorAdd -----}
procedure VectorAdd(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var nX, nY, nZ : double);
begin
  nX := v1X + v2X;
  nY := v1Y + v2Y;
  nZ := v1Z + v2Z;
end;

{----- VectorAdd -----}
procedure VectorAdd(const v1, v2 : TVector3f; var v3 : TVector3f);
begin
  VectorAdd(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, v3.X, v3.Y, v3.Z);
end;

{----- VectorAdd -----}
function VectorAdd(const v1, v2 : TVector3f) : TVector3f;
begin
  VectorAdd(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, result.X, result.Y, result.Z);
end;

{----- VectorSubtract -----}
procedure VectorSubtract(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var nX, nY, nZ : double);
begin
  nX := v1X - v2X;
  nY := v1Y - v2Y;
  nZ := v1Z - v2Z;
end;

{----- VectorSubtract -----}
procedure VectorSubtract(const v1, v2 : TVector3f; var v3 : TVector3f);
begin
  VectorSubtract(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, v3.X, v3.Y, v3.Z);
end;

{----- VectorSubtract -----}
function VectorSubtract(const v1, v2 : TVector3f) : TVector3f;
begin
  VectorSubtract(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, result.X, result.Y, result.Z);
end;

{----- VectorNegate -----}
procedure VectorNegate(const vX, vY, vZ : double; var nX, nY, nZ : double);
begin
  nX := -vX;
  nY := -vY;
  nZ := -vZ;
end;

{----- VectorNegate -----}
procedure VectorNegate(const v1 : TVector3f; var v2 : TVector3f);
begin
  VectorNegate(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z);
end;

{----- VectorNegate -----}
function VectorNegate(const v : TVector3f) : TVector3f;
begin
  VectorNegate(v.X, v.Y, v.Z, result.X, result.Y, result.Z);
end;

{----- VectorCrossProduct -----}
procedure VectorCrossProduct(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var nX, nY, nZ : double);
begin
	nX := (v1Y*v2Z) - (v1Z*v2Y);
	nY := (v1Z*v2X) - (v1X*v2Z);
	nZ := (v1X*v2Y) - (v1Y*v2X);
end;

{----- VectorCrossProduct -----}
procedure VectorCrossProduct(const v1, v2 : TVector3f; var v3 : TVector3f);
begin
  VectorCrossProduct(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, v3.X, v3.Y, v3.Z);
end;

{----- VectorCrossProduct -----}
function VectorCrossProduct(const v1, v2 : TVector3f) : TVector3f;
begin
  VectorCrossProduct(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, result.X, result.Y, result.Z);
end;

{----- ProjectPoint -----}
procedure ProjectPoint(const Srcx, Srcy, Srcz, Dstx, Dsty, Dstz, Dist : double; out Nx, Ny, Nz : double);
var
  DistRatio : double;
begin
  if SameValue(VectorDistance(Srcx,Srcy,Srcz,Dstx,Dsty,Dstz), 0) then
  begin
    Nx := Srcx;
    Ny := Srcy;
    Nz := Srcz;
  end
  else
  begin
    DistRatio := Dist / VectorDistance(Srcx, Srcy, Srcz, Dstx, Dsty, Dstz);
    Nx := Srcx + DistRatio * (Dstx - SrcX);
    Ny := Srcy + DistRatio * (Dsty - Srcy);
    Nz := Srcz + DistRatio * (Dstz - Srcz);
  end;
end;

{----- ProjectPoint -----}
procedure ProjectPoint(const Src, Dst : TVector3f; const Dist : double; out vOut : TVector3f);
begin
  ProjectPoint(Src.X, Src.Y, Src.Z, Dst.X, Dst.Y, Dst.Z, Dist, vOut.X, vOut.Y, vOut.Z);
end;

{----- VectorRotate -----}
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const x, y, z : double; out Nx, Ny, Nz : double);
var
  TempX : double;
  TempY : double;
  TempZ : double;
  SinX  : extended;
  SinY  : extended;
  SinZ  : extended;
  CosX  : extended;
  CosY  : extended;
  CosZ  : extended;
begin
  SinX := Sin(angleRadY);
  SinY := Sin(angleRadX);
  SinZ := Sin(angleRadZ);
  CosX := Cos(angleRadY);
  CosY := Cos(angleRadX);
  CosZ := Cos(angleRadZ);

  Tempy := y * CosY -     z * SinY;
  Tempz := y * SinY +     z * CosY;
  Tempx := x * CosX - Tempz * SinX;

  Nz :=     x * SinX + Tempz * CosX;
  Nx := Tempx * CosZ - TempY * SinZ;
  Ny := Tempx * SinZ + TempY * CosZ;
end;

{----- VectorRotate -----}
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const vIn : TVector3f; out vOut : TVector3f);
begin
  VectorRotate(angleRadX, angleRadY, angleRadZ, vIn.X, vIn.Y, vIn.Z, vOut.X, vOut.Y, vOut.Z);
end;

{----- VectorRotate -----}
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const x, y, z, Ox, Oy, Oz : double; out Nx, Ny, Nz : double);
begin
  VectorRotate(angleRadX, angleRadY, angleRadZ, x - Ox, y - Oy, z - Oz, Nx, Ny, Nz);
  Nx := Nx + Ox;
  Ny := Ny + Oy;
  Nz := Nz + Oz;
end;

{----- VectorRotate -----}
procedure VectorRotate(const angleRadX, angleRadY, angleRadZ : double; const vIn, vAround : TVector3f; out vOut : TVector3f);
begin
  VectorRotate(angleRadX, angleRadY, angleRadZ, vIn.X, vIn.Y, vIn.Z, vAround.X, vAround.Y, vAround.Z, vOut.X, vOut.Y, vOut.Z);
end;

{----- RotateVector -----}
procedure RotateVector(var nX, nY, nZ : double; const axisX, axisY, axisZ, angleRad : double);
var
  mX_X, mX_Y, mX_Z, mX_W : double;
  mY_X, mY_Y, mY_Z, mY_W : double;
  mZ_X, mZ_Y, mZ_Z, mZ_W : double;
  mW_X, mW_Y, mW_Z, mW_W : double;
  outX, outY, outZ : double;
begin
  CreateRotationMatrix(axisX, axisY, axisZ, angleRad, mX_X, mX_Y, mX_Z, mX_W, mY_X, mY_Y, mY_Z, mY_W, mZ_X, mZ_Y, mZ_Z, mZ_W, mW_X, mW_Y, mW_Z, mW_W);
  VectorMatrixTransform(nX, nY, nZ, mX_X, mX_Y, mX_Z, mX_W, mY_X, mY_Y, mY_Z, mY_W, mZ_X, mZ_Y, mZ_Z, mZ_W, mW_X, mW_Y, mW_Z, mW_W, outX, outY, outZ);
  nX := outX;
  nY := outY;
  nZ := outZ;
end;

{----- RotateVector -----}
procedure RotateVector(var vIn : TVector3f; const vAxis : TVector3f; const angleRad : double);
begin
  RotateVector(vIn.X, vIn.Y, vIn.Z, vAxis.X, vAxis.Y, vAxis.Z, angleRad);
end;

{----- Lerp -----}
function Lerp(const start, stop, t : double) : double;
begin
  Result := start + (stop - start) * t;
end;

{----- AngleLerp -----}
function AngleLerp(startRad, stopRad : double; const t : double) : double;
var
  d : double;
begin
  startRad := NormalizeAngle(startRad);
  stopRad  := NormalizeAngle(stopRad);
  d        := stopRad - startRad;
       if d>+PI then d := -d - c2PI
  else if d<-PI then d :=  d + c2PI;
  Result := startRad + d * t;
end;

{----- Quadrant -----}
function Quadrant(const angle : double) : integer;
begin
  result := 0;
       if (angle >=   0.0) and (angle <  90.0) then result := 1
  else if (angle >=  90.0) and (angle < 180.0) then result := 2
  else if (angle >= 180.0) and (angle < 270.0) then result := 3
  else if (angle >= 270.0) and (angle < 360.0) then result := 4
  else if angle = 360.0                        then result := 1;
end;

{----- CartesianAngle -----}
function CartesianAngle(const x, y : double) : double;
begin
  if      (x > 0.0) and (y > 0.0) then Result := (ArcTan( y / x) * c180DivPI)
  else if (x < 0.0) and (y > 0.0) then Result := (ArcTan(-x / y) * c180DivPI) +  90.0
  else if (x < 0.0) and (y < 0.0) then Result := (ArcTan( y / x) * c180DivPI) + 180.0
  else if (x > 0.0) and (y < 0.0) then Result := (ArcTan(-x / y) * c180DivPI) + 270.0
  else if (x = 0.0) and (y > 0.0) then Result :=  90.0
  else if (x < 0.0) and (y = 0.0) then Result := 180.0
  else if (x = 0.0) and (y < 0.0) then Result := 270.0
  else                                 Result := 0.0;
end;

{----- CartesianPoint -----}
procedure CartesianPoint(const angle, radius : double; var x, y : double);
begin
       if angle =   0.0 then begin x :=  radius; y := 0;       end
  else if angle =  90.0 then begin x := 0;       y :=  radius; end
  else if angle = 180.0 then begin x := -radius; y := 0;       end
  else if angle = 270.0 then begin x := 0;       y := -radius; end
  else if angle = 360.0 then begin x :=  radius; y := 0;       end
  else if (angle > 0) and (angle < 90) then
  begin
    x := radius * cos(DegToRad(angle));
    y := radius * sin(DegToRad(angle));
  end
  else if (angle > 90) and (angle < 180) then
  begin
    x := -radius * sin(DegToRad(angle - 90));
    y :=  radius * cos(DegToRad(angle - 90));
  end
  else if (angle > 180) and (angle < 270) then
  begin
    x := -radius * cos(DegToRad(angle - 180));
    y := -radius * sin(DegToRad(angle - 180));
  end
  else if (angle > 270) and (angle < 360) then
  begin
    x :=  radius * sin(DegToRad(angle - 270));
    y := -radius * cos(DegToRad(angle - 270));
  end;
end;

{----- NormalizeAngle -----}
function NormalizeAngle(const angleRad : double) : double;
begin
  Result := angleRad - Int(angleRad * cInv2PI) * c2PI;
       if Result>PI  then Result := Result - 2*PI
  else if Result<-PI then Result := Result + 2*PI;
end;

{----- VectorLerp -----}
procedure VectorLerp(const v1X, v1Y, v1Z, v2X, v2Y, v2Z, t : double; var vrX, vrY, vrZ : double);
begin
  vrX := v1X + (v2X - v1X) * t;
  vrY := v1Y + (v2Y - v1Y) * t;
  vrZ := v1Z + (v2Z - v1Z) * t;
end;

{----- VectorLerp -----}
procedure VectorLerp(const v1, v2 : TVector3f; const t : double; var vOut : TVector3f);
begin
  VectorLerp(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, t, vOut.X, vOut.Y, vOut.Z);
end;

{----- VectorLerp -----}
function VectorLerp(const v1, v2 : TVector3f; const t : double) : TVector3f;
begin
  VectorLerp(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, t, result.X, result.Y, result.Z);
end;

{----- AngleBetweenTwoVectors -----}
function AngleBetweenTwoVectors(x1, y1, z1, x2, y2, z2, x3, y3, z3 : double) : double;
var
  ab : double;
  Dist : double;
begin
  result := 0;
  if VectorEquals(x1, y1, z1, x2, y2, z2) or
     VectorEquals(x2, y2, z2, x3, y3, z3) or
     VectorEquals(x1, y1, z1, x3, y3, z3) then exit;

  x1 := x1 - x2;
  x3 := x3 - x2;
  y1 := y1 - y2;
  y3 := y3 - y2;
  z1 := z1 - z2;
  z3 := z3 - z2;

  //Calculate Lay Distance
  Dist := (x1 * x1 + y1 * y1 + z1 * z1) * (x3 * x3 + y3 * y3 + z3 * z3);

  if Dist = 0 then Result := 0.0
  else
  begin
    ab := (x1 * x3 + y1 * y3 + z1 * z3) / sqrt(Dist);
    result := ArcCos(ab) * c180DivPI;

    // possible ArcCos values -1 / +1
    if IsNaN(result) then
      if ab > 0 then result := 0
                else result := 180;
  end;
end;

{----- AngleBetweenTwoVectors -----}
function AngleBetweenTwoVectors(v1, v2, v3 : TVector3f) : double;
begin
  Result := AngleBetweenTwoVectors(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, v3.X, v3.Y, v3.Z);
end;

{----- CreateRotationMatrix -----}
procedure CreateRotationMatrix(const axisX, axisY, axisZ, angleRad : double; var mX_X, mX_Y, mX_Z, mX_W, mY_X, mY_Y, mY_Z, mY_W, mZ_X, mZ_Y, mZ_Z, mZ_W, mW_X, mW_Y, mW_Z, mW_W : double);
var
  aX : double;
  aY : double;
  aZ : double;
  cosine : double;
  sine : double;
  one_minus_cosine : double;
begin
  SinCos(angleRad, sine, cosine);
  one_minus_cosine := 1-cosine;
  VectorNormalize(axisX, axisY, axisZ, aX, aY, aZ);

  mX_X := (one_minus_cosine * aX * aX) + cosine;
  mX_Y := (one_minus_cosine * aX * aY) - (aZ * sine);
  mX_Z := (one_minus_cosine * aZ * aX) + (aY * sine);
  mX_W := 0;

  mY_X := (one_minus_cosine * aX * aY) + (aZ * sine);
  mY_Y := (one_minus_cosine * aY * aY) + cosine;
  mY_Z := (one_minus_cosine * aY * aZ) - (aX * sine);
  mY_W := 0;

  mZ_X := (one_minus_cosine * aZ * aX) - (aY * sine);
  mZ_Y := (one_minus_cosine * aY * aZ) + (aX * sine);
  mZ_Z := (one_minus_cosine * aZ * aZ) + cosine;
  mZ_W := 0;

  mW_X := 0;
  mW_Y := 0;
  mW_Z := 0;
  mW_W := 1;
end;

{----- CreateRotationMatrix -----}
procedure CreateRotationMatrix(const axisVec : TVector3f; const angleRad : double; var matrixOut : TMatrix16f);
begin
  CreateRotationMatrix(axisVec.X, axisVec.Y, axisVec.Z, angleRad, matrixOut.X1, matrixOut.Y1, matrixOut.Z1, matrixOut.W1, matrixOut.X2, matrixOut.Y2, matrixOut.Z2, matrixOut.W2, matrixOut.X3, matrixOut.Y3, matrixOut.Z3, matrixOut.W3, matrixOut.X4, matrixOut.Y4, matrixOut.Z4, matrixOut.W4);
end;

{----- VectorMatrixTransform -----}
procedure VectorMatrixTransform(const vX, vY, vZ, mX_X, mX_Y, mX_Z, mX_W, mY_X, mY_Y, mY_Z, mY_W, mZ_X, mZ_Y, mZ_Z, mZ_W, mW_X, mW_Y, mW_Z, mW_W : double; var nX, nY, nZ : double);
begin
  nX := vX * mX_X + vY * mY_X + vZ * mZ_X + mW_X;
  nY := vX * mX_Y + vY * mY_Y + vZ * mZ_Y + mW_Y;
  nZ := vX * mX_Z + vY * mY_Z + vZ * mZ_Z + mW_Z;
end;

{----- VectorMatrixTransform -----}
procedure VectorMatrixTransform(const Vec : TVector3f; const Mat : TMatrix16f; var vOut : TVector3f);
begin
  VectorMatrixTransform(Vec.X, Vec.Y, Vec.Z, Mat.X1, Mat.Y1, Mat.Z1, Mat.W1, Mat.X2, Mat.Y2, Mat.Z2, Mat.W2, Mat.X3, Mat.Y3, Mat.Z3, Mat.W3, Mat.X4, Mat.Y4, Mat.Z4, Mat.W4, vOut.X, vOut.Y, vOut.Z);
end;

{----- VectorDotProduct -----}
function VectorDotProduct(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double) : double;
begin
  Result := v1X * v2X + v1Y * v2Y + v1Z * v2Z;
end;

{----- VectorDotProduct -----}
procedure VectorDotProduct(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var dot : double);
begin
  dot := VectorDotProduct(v1X, v1Y, v1Z, v2X, v2Y, v2Z);
end;

{----- VectorDotProduct -----}
function VectorDotProduct(const v1, v2 : TVector3f) : double;
begin
  Result := VectorDotProduct(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z);
end;

{----- VectorDotProduct -----}
procedure VectorDotProduct(const v1, v2 : TVector3f; var dot : double);
begin
  dot := VectorDotProduct(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z);
end;

{----- DirectionVector -----}
procedure DirectionVector(const v1X, v1Y, v1Z, v2X, v2Y, v2Z : double; var vrX, vrY, vrZ : double);
begin
  VectorNormalize(v2X - v1X, v2Y - v1Y, v2Z - v1Z, vrX, vrY, vrZ);
end;

{----- DirectionVector -----}
procedure DirectionVector(const v1, v2 : TVector3f; var vOut : TVector3f);
begin
  DirectionVector(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, vOut.X, vOut.Y, vOut.Z);
end;

{----- DirectionVector -----}
function DirectionVector(const v1, v2 : TVector3f) : TVector3f;
begin
  DirectionVector(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, result.X, result.Y, result.Z);
end;

{----- VectorCombinePosRotSca -----}
function VectorCombinePosRotSca(const v, position, rotation, scale : TVector3f) : TVector3f;
var
  vTmp : TVector3f;
begin
  vTmp := MakeVector3f(v.X * scale.X + position.X, v.Y * scale.Y + position.Y, v.Z * scale.Z + position.Z);
  VectorRotate(DegToRad(rotation.X), DegToRad(rotation.Y), DegToRad(rotation.Z), vTmp.X, vTmp.Y, vTmp.Z, position.X, position.Y, position.Z, vTmp.X, vTmp.Y, vTmp.Z);
  result := vTmp;
end;

{----- VectorCombinePosRotSca -----}
function VectorCombinePosRotSca(const v, position, rotationX, rotationY, rotationZ, scale : TVector3f) : TVector3f;
var
  vTmp : TVector3f;
begin
  vTmp := MakeVector3f(v.X * scale.X + position.X, v.Y * scale.Y + position.Y, v.Z * scale.Z + position.Z);
  VectorRotate(DegToRad(rotationX.X), DegToRad(rotationX.Y), DegToRad(rotationX.Z), vTmp.X, vTmp.Y, vTmp.Z, position.X, position.Y, position.Z, vTmp.X, vTmp.Y, vTmp.Z);
  VectorRotate(DegToRad(rotationY.X), DegToRad(rotationY.Y), DegToRad(rotationY.Z), vTmp.X, vTmp.Y, vTmp.Z, position.X, position.Y, position.Z, vTmp.X, vTmp.Y, vTmp.Z);
  VectorRotate(DegToRad(rotationZ.X), DegToRad(rotationZ.Y), DegToRad(rotationZ.Z), vTmp.X, vTmp.Y, vTmp.Z, position.X, position.Y, position.Z, vTmp.X, vTmp.Y, vTmp.Z);
  result := vTmp;
end;

{----- VectorSide_3D -----}
function VectorSide_3D(const x1, y1, z1, x2, y2, z2, x3, y3, z3, Px, Py, Pz : double; const epsilon : double = 0.00001) : integer;
var
  Px1  : double;
  Px2  : double;
  Px3  : double;
  Py1  : double;
  Py2  : double;
  Py3  : double;
  Pz1  : double;
  Pz2  : double;
  Pz3  : double;
  Orin : double;
begin
  Px1 := x1 - px;
  Px2 := x2 - px;
  Px3 := x3 - px;

  Py1 := y1 - py;
  Py2 := y2 - py;
  Py3 := y3 - py;

  Pz1 := z1 - pz;
  Pz2 := z2 - pz;
  Pz3 := z3 - pz;

  Orin := Px1 * (Py2 * Pz3 - Pz2 * Py3) + Px2 * (Py3 * Pz1 - Pz3 * Py1) + Px3 * (Py1 * Pz2 - Pz1 * Py2);

  if SameValue(Orin, 0, epsilon) then result := 0  // vector lays in same plane
  else if Orin < 0 then result := +1               // vector lays on first side from plane
                   else result := -1;              // vector lays on second side from plane
end;

{----- VectorSide_3D -----}
function VectorSide_3D(const v1, v2, v3, vP : TVector3f; const epsilon : double = 0.00001) : integer;
begin
  result := VectorSide_3D(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, v2.X, v3.Y, v3.Z, vP.X, vP.Y, vP.Z, epsilon);
end;

{----- CollinearTest -----}
function CollinearTest(const x1, y1, z1, x2, y2, z2, x3, y3, z3 : double; const epsilon : double = 0.00001) : boolean;
var
  Dx1 : double;
  Dx2 : double;
  Dy1 : double;
  Dy2 : double;
  Dz1 : double;
  Dz2 : double;
  Cx  : double;
  Cy  : double;
  Cz  : double;
begin
  Dx1 := x2 - x1;
  Dy1 := y2 - y1;
  Dz1 := z2 - z1;

  Dx2 := x3 - x1;
  Dy2 := y3 - y1;
  Dz2 := z3 - z1;

  Cx  := (Dy1 * Dz2) - (Dy2 * Dz1);
  Cy  := (Dx2 * Dz1) - (Dx1 * Dz2);
  Cz  := (Dx1 * Dy2) - (Dx2 * Dy1);

  Result := SameValue(Cx * Cx + Cy * Cy + Cz * Cz, 0.0, epsilon);
end;

{----- CollinearTest -----}
function CollinearTest(const v1, v2, v3 : TVector3f; const epsilon : double = 0.00001) : boolean;
begin
  result := CollinearTest(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, v3.X, v3.Y, v3.Z, epsilon);
end;

{----- CoplanarTest -----}
function CoplanarTest(const vArr : TVectorArray3f; const epsilon : double = 0.00001) : boolean;
var
  i : integer;
begin
  result := false;

  // minimal number of vectors in array for testing
  if length(vArr) < 4 then exit;

  for i := 3 to length(vArr) - 1 do
    if not VectorSide_3D(vArr[0].X, vArr[0].Y, vArr[0].Z,
                         vArr[1].X, vArr[1].Y, vArr[1].Z,
                         vArr[2].X, vArr[2].Y, vArr[2].Z,
                         vArr[i].X, vArr[i].Y, vArr[i].Z, epsilon) = 0 then exit;
  result := true;
end;

{----- CoplanarTest -----}
function CoplanarTest(const vPlane, vArr : TVectorArray3f; const epsilon : double = 0.00001) : boolean;
var
  i : integer;
begin
  result := false;

  // minimal number of vectors in array for testing
  if length(vPlane) < 4 then exit;

  for i := 3 to length(vPlane) - 1 do
    if not VectorSide_3D(vPlane[0].X, vPlane[0].Y, vPlane[0].Z,
                         vPlane[1].X, vPlane[1].Y, vPlane[1].Z,
                         vPlane[2].X, vPlane[2].Y, vPlane[2].Z,
                         vPlane[i].X, vPlane[i].Y, vPlane[i].Z, epsilon) = 0 then exit;

  for i := 0 to length(vArr) - 1 do
    if not VectorSide_3D(vPlane[0].X, vPlane[0].Y, vPlane[0].Z,
                         vPlane[1].X, vPlane[1].Y, vPlane[1].Z,
                         vPlane[2].X, vPlane[2].Y, vPlane[2].Z,
                         vArr[i].X, vArr[i].Y, vArr[i].Z, epsilon) = 0 then exit;
  result := true;
end;

{----- VectorInVectorArrayExists -----}
function VectorInVectorArrayExists(const v : TVector3f; const Arr : TVectorArray3f; var arr_index : integer; const epsilon : double = 0.00001) : boolean;
var
  i : integer;
begin
  result := false;
  for i := 0 to length(Arr) - 1 do
    if VectorEquals(Arr[i], v, epsilon) then
    begin
      arr_index := i;
      result := true;
      break;
    end;
end;

{----- VectorInVectorArrayArrayExists -----}
function VectorInVectorArrayArrayExists(const v : TVector3f; const Arr : TVectorArrayArray3f; var first_arr_index, second_arr_index : integer; const epsilon : double = 0.00001) : boolean;
var
  i : integer;
  a : integer;
begin
  result := false;
  for i := 0 to length(Arr) - 1 do
    for a := 0 to length(Arr[i]) - 1 do
      if VectorEquals(Arr[i, a], v, epsilon) then
      begin
        first_arr_index := i;
        second_arr_index := a;
        result := true;
        exit;
      end;
end;

end.
